'''
This module contains the menu bar command classes, that inherit from
tk_steroids' MenuMaker for easy initialization.

In the beginning of the module, there are some needed functions.
'''


import os
import subprocess

import numpy as np
import tkinter as tk
import tkinter.messagebox as messagebox
import tkinter.filedialog as filedialog
import tkinter.simpledialog as simpledialog

from tk_steroids.elements import TickboxFrame
from tk_steroids.dialogs import popup_tickselect, popup
from tk_steroids.menumaker import MenuMaker
from tk_steroids.datamanager import ListManager

import gonioanalysis
from gonioanalysis.directories import USER_HOMEDIR, ANALYSES_SAVEDIR
from gonioanalysis.droso import SpecimenGroups
from gonioanalysis.drosom import linked_data
from gonioanalysis.drosom import kinematics
from gonioanalysis.drosom import sinesweep
from gonioanalysis.drosom import export
from gonioanalysis.drosom.reports.left_right import left_right_displacements, lrfiles_summarise
from gonioanalysis.drosom.reports.stats import response_magnitudes
from gonioanalysis.tkgui import settings
from gonioanalysis.tkgui.run_measurement import MeasurementWindow
from gonioanalysis.tkgui.widgets import (
        select_specimens,
        select_specimen_groups,
        ZeroCorrect,
        CompareVectormaps,
        ImagefolderMultisel,
        ExportWizard,
        )




def ask_string(title, prompt, tk_parent):
    '''
    Asks the user for a string.
    '''
    string = simpledialog.askstring(title, prompt, parent=tk_parent)
    return string



def prompt_result(tk_root, string, title='Message'):
    '''
    Shows the result and also sets it to the clipboard
    '''
    tk_root.clipboard_clear()
    tk_root.clipboard_append(string)

    messagebox.showinfo(
            title=title,
            message=f'Following has been copied to your clipboard:\n\n{string}')



class ModifiedMenuMaker(MenuMaker):
    '''
    Modify the MenuMaker so that at object initialization, we pass
    the common core to all of the command objects.
    '''

    def __init__(self, tk_root, core, *args, **kwargs):
        '''
        core        An instance of the core.py Core Class
        '''
        super().__init__(*args, **kwargs)
        self.core = core
        self.tk_root = tk_root

        self.replacement_dict['DASH'] = '-'
        self.replacement_dict['_'] = ' '
    
    def _message(self, message, **kwargs):
        if message == 'nospecimen':
            message = 'Select a specimen first'

        prompt_result(self.tk_root, message, **kwargs)
    
    def _ask_string(self, message, title='Input text'):
        return ask_string(title, message, self.tk_root)


class FileCommands(ModifiedMenuMaker):
    '''
    File menu commands for examine view.
    '''

    def _force_order(self):
        '''
        Let's force menu ordering. See the documentation from the
        menu maker.
        '''
        menu = ['set_data_directory',
                'add_data_directory',
                '.',
                'create_virtual_analyser',
                '.',
                'settings',
                '.',
                'exit']
        return menu


    def set_data_directory(self, append=False):
        '''
        Asks user for the data directory and sets it active in Core.
        '''
        previousdir = settings.get('last_datadir', default=USER_HOMEDIR)
        
        # Check if the previous data directory stil exists (usb drive for example)
        if not os.path.isdir(previousdir):
            previousdir = USER_HOMEDIR 

        directory = filedialog.askdirectory(
                parent=self.tk_root,
                title='Select directory containing specimens',
                mustexist=True,
                initialdir=previousdir
                )

        if not directory:
            return None
        
        if append == False:
            self.core.set_data_directory([directory])
        else:
            self.core.set_data_directory(self.core.data_directory + [directory])
        self.core.update_gui(changed_specimens=True)
        
        settings.set('last_datadir', directory)


    def add_data_directory(self):
        '''
        Like set_data_directory, but instead of changing the data directory,
        comdines the entries from previous and the new data directories.
        '''
        self.set_data_directory(append=True)


    def create_virtual_analyser(self):
        
        def create_virt(specimens):
            self.core.create_virtual_analyser('VirtTest', specimens)
    
        select_specimens(self.core, create_virt, with_movements=True)

    def settings(self):
        pass


    def exit(self):
        self.tk_root.winfo_toplevel().destroy()




class ImageFolderCommands(ModifiedMenuMaker):
    '''
    Commands for the currently selected image folder.
    '''
   
    def _force_order(self):
        return ['select_ROIs', 'measure_movement',
                'measure_movement_DASH_in_absolute_coordinates',
                '.',
                'max_of_the_mean_response',
                'half_rise_time',
                'half_rise_time_DASH_fit_to_mean',
                'latency',
                'latency_DASH_fit_to_mean',
                '.',
                'start_position_analysis',
                '.',
                'open_in_ImageJ']

    def max_of_the_mean_response(self):
        
        result = kinematics.mean_max_response(self.core.analyser, self.core.selected_recording)
        prompt_result(self.tk_root, result, 'Max of the mean (pixels)')
    

    def half_rise_time(self, fit_to_mean=False):
        result = kinematics.sigmoidal_fit(
                self.core.analyser, self.core.selected_recording,
                fit_to_mean=fit_to_mean)[2]
        prompt_result(self.tk_root, str(np.mean(result)), 'Half-rise time (s)')
    

    def half_rise_time_DASH_fit_to_mean(self):
        self.half_rise_time(fit_to_mean=True)

    def latency(self, fit_to_mean=False):
        result = kinematics.latency(
                self.core.analyser, self.core.selected_recording,
                fit_to_mean=fit_to_mean)
        prompt_result(self.tk_root, str(np.mean(result)), 'Latency (s)')

    def latency_DASH_fit_to_mean(self):
        self.latency(True)

    def select_ROIs(self):
        self.core.analyser.select_ROIs(callback_on_exit=self.core.update_gui,
                reselect_fns=[self.core.selected_recording], old_markings=True)


    def measure_movement(self, absolute_coordinates=False):
        '''
        Run Movemeter (cross-correlation) on the selected image folder.
        '''
        func = lambda stop: self.core.analyser.measure_both_eyes(only_folders=str(self.core.selected_recording), absolute_coordinates=absolute_coordinates, stop_event=stop)
        
        MeasurementWindow(self.tk_root, [func], title='Measure movement', callback_on_exit=lambda: self.core.update_gui(changed_specimens=True))


    def measure_movement_DASH_in_absolute_coordinates(self):
        self.measure_movement(absolute_coordinates=True)
    

    def open_in_ImageJ(self):
        '''Opens the currently selected image folder in ImageJ
        '''
        folder = self.core.selected_recording
        fns = self.core.analyser.list_images(folder, absolute_path=True)
        subprocess.Popen(['imagej', *fns,])
    

    def start_position_analysis(self):
        self.core.adm_subprocess(
                'current',
                f'-A startpos --analysis-options "image_folder={self.core.selected_recording}"')



class SpecimenCommands(ModifiedMenuMaker):
    '''
    Commands for the currentlt selected specimen.
    '''

    def _force_order(self):
        return ['set_active_analysis',
                'set_vector_rotation_offset',
                'set_vertical_zero_rotation',
                'set_yaw_rotation',
                '.',
                'select_ROIs',
                'measure_movement',
                'measure_movement_DASH_in_absolute_coordinates',
                '.',
                'mean_displacement_over_time',
                'latency_by_sigmoidal_fit',
                '.',
                'vectormap_DASH_interactive_plot',
                'vectormap_DASH_rotating_video',
                'vectormap_DASH_export',
                '.',
                'start_position_analysis',
                '.']


    def set_active_analysis(self):

        name = ask_string('Active analysis', 'Give new or existing analysis name (empty for default)', self.tk_root)
        
        self.core.active_analysis = name
        if self.core.analyser:
            self.core.analyser.active_analysis = name
        self.tk_root.status_active_analysis.config(text='Active analysis: {}'.format(self.core.active_analysis))
        
        self.core.update_gui(changed_specimens=True) 


    def set_vector_rotation_offset(self):

        rotation = ask_string('Active analysis', 'Give new or existing analysis name (empty for default)', self.tk_root)
        
        if rotation:
            self.core.analyser.vector_rotation = float(rotation)
        else:
            self.core.analyser.vector_rotation = None

        

    def select_ROIs(self):
        '''
        Select regions of interests (ROIs) for the currently selected specimen.
        '''
            
        # Ask confirmation if ROIs already selected
        if self.core.analyser.are_rois_selected():
            sure = messagebox.askokcancel('Reselect ROIs', 'Are you sure you want to reselect ROIs?')
            if not sure:
                return None
       
        self.core.analyser.select_ROIs(callback_on_exit=lambda: self.core.update_gui(changed_specimens=True))



    def measure_movement(self, absolute_coordinates=False):
        '''
        Run Movemeter (cross-correlation) on the specimen.
        '''
        if not self.core.current_specimen:
            self._message('nospecimen')
            return None

        # Ask confirmation if ROIs already selected
        if self.core.analyser.is_measured():
            sure = messagebox.askokcancel('Remeasure movements', 'Are you sure you want to remeasure?')
            if not sure:
                return None
        
        func = lambda stop: self.core.analyser.measure_both_eyes(absolute_coordinates=absolute_coordinates, stop_event=stop)
        
        if self.core.analyser.__class__.__name__ != 'OAnalyser':
            MeasurementWindow(self.tk_root, [func], title='Measure movement', callback_on_exit=lambda: self.core.update_gui(changed_specimens=True))
        else:
            # OAnalyser; Threading in MeasurementWindow would cause problems for plotting
            func(stop=None)
            self.core.update_gui(changed_specimens=True)


    def measure_movement_DASH_in_absolute_coordinates(self):
        self.measure_movement(absolute_coordinates=True)


    def set_vertical_zero_rotation(self):
        '''
        Start antenna level search for the current specimen (zero correction)
        '''
        
        # Try to close and destroy if any other antenna_level
        # windows are open (by accident)
        try:
            self.correct_window.destroy()
        except:
            # Nothing to destroy
            pass
        
        if not self.core.current_specimen:
            self._message("nospecimen")
        else:
            
            self.correct_window = tk.Toplevel()
            self.correct_window.title('Zero correction -  {}'.format(self.core.current_specimen))
            self.correct_window.grid_columnconfigure(0, weight=1)
            self.correct_window.grid_rowconfigure(0, weight=1)

            def callback():
                self.correct_window.destroy()
                self.core.update_gui()

            self.correct_frame = ZeroCorrect(self.correct_window,
                    self.core.get_specimen_fullpath(),
                    'alr_data',
                    callback=callback)
            self.correct_frame.grid(sticky='NSEW')


    def set_yaw_rotation(self):
        # Try to close and destroy if any other antenna_level
        # windows are open (by accident)
        try:
            self.yaw_window.destroy()
        except:
            # Nothing to destroy
            pass
        
        if not self.core.current_specimen:
            self._message("nospecimen")
        else:
            
            self.yaw_window = tk.Toplevel()
            self.yaw_window.title(f'Yaw rotation set -  {self.core.current_specimen}')
            self.yaw_window.grid_columnconfigure(0, weight=1)
            self.yaw_window.grid_rowconfigure(0, weight=1)

            def on_yaw():
                
                yaw = self.yaw_frame.ticked[0]
                self.core.analyser.attributes['yaw'] = int(yaw)
                self.core.analyser.save_attributes()

                self.yaw_window.destroy()
                self.core.update_gui()

            yaws = ['-90','0','90']
            yaw = str(self.core.analyser.attributes['yaw'])
            defaults = [False,False,False]
            defaults[yaws.index(yaw)] = True

            self.yaw_frame = TickboxFrame(
                    self.yaw_window, yaws, defaults=defaults,
                    single_select=True)
            self.yaw_frame.grid(
                    row=0, column=0, columnspan=2,
                    sticky='NSEW')
            
            ok = tk.Button(self.yaw_window, text='Ok', command=on_yaw)
            ok.grid(row=1, column=1)
            
            cancel = tk.Button(self.yaw_window, text='Cancel',
                               command=self.yaw_window.destroy)
            cancel.grid(row=1, column=0)

    def vectormap_DASH_interactive_plot(self):
        self.core.adm_subprocess('current', '-A vectormap')


    def vectormap_DASH_rotating_video(self):
        self.core.adm_subprocess('current', '--tk_waiting_window -A vectormap_video')

    
    def vectormap_DASH_export(self):
        analysername = self.core.analyser.get_specimen_name()
        fn = tk.filedialog.asksaveasfilename(initialfile=analysername, defaultextension='.npy', filetypes=export.FILETYPES)
        if fn:
            self.core.adm_subprocess('current', f'-A export_vectormap --output "{fn}"')
            
    
    def mean_displacement_over_time(self):
        self.core.adm_subprocess('current', '-A magtrace --analysis-options mean_repeats=True milliseconds=True mean_imagefolders=True')


    def latency_by_sigmoidal_fit(self):
        results_string = ''
        for image_folder in self.core.analyser.list_imagefolders():
            result = kinematics.sigmoidal_fit(self.core.analyser, image_folder)[2]
            if result is not None:
                result = np.mean(result)
            results_string += '{}   {}\n'.format(image_folder, result)
        

        prompt_result(self.tk_root, results_string, 'Mean latency (s)')

    def start_position_analysis(self):
        self.core.adm_subprocess('current', '-A startpos')


class ManySpecimenCommands(ModifiedMenuMaker):
    '''
    Commands for all of the specimens in the current data directory.
    Usually involves a checkbox to select the wanted specimens.
    '''

    def _force_order(self):
        return ['measure_movements_DASH_list_all', 'measure_movements_DASH_list_only_unmeasured',
                'measure_movements_DASH_in_absolute_coordinates',
                '.',
                'averaged_vectormap_DASH_interactive_plot', 'averaged_vectormap_DASH_rotating_video',
                'averaged_vectormap_DASH_rotating_video_DASH_set_title',
                'averaged_vectormap_DASH_export',
                '.',
                'compare_vectormaps',
                '.',
                'comparision_to_optic_flow_DASH_video',
                '.',
                'export_LR_displacement_CSV',
                'export_LR_displacement_CSV_DASH_strong_weak_eye_division',
                'save_kinematics_analysis_CSV',
                'save_kinematics_analysis_CSV_DASH_fit_to_mean',
                'save_sinesweep_analysis_CSV',
                '.',
                'export_wizard',
                '.',
                ]


    def _batch_measure(self, specimens, absolute_coordinates=False):
        
        # Here lambda requires specimen=specimen keyword argument; Otherwise only
        # the last specimen gets analysed N_specimens times
        targets = [lambda stop, specimen=specimen: self.core.get_manalyser(specimen).measure_both_eyes(absolute_coordinates=absolute_coordinates, stop_event=stop) for specimen in specimens]
    

        if self.core.analyser_class.__name__ != 'OAnalyser':
            MeasurementWindow(self.parent_menu.winfo_toplevel(), targets, title='Measure movement',
                    callback_on_exit=lambda: self.core.update_gui(changed_specimens=True))
        else:
            # For OAnalyser; Threading in MeasurementWindow causes problems for plotting
            for target in targets:
                target()
            self.core.update_gui(changed_specimens=True)


    def measure_movements_DASH_list_all(self):

        select_specimens(self.core, self._batch_measure, with_rois=True)


    def measure_movements_DASH_list_only_unmeasured(self):

        select_specimens(self.core, self._batch_measure, with_rois=True, with_movements=False)

    
    def measure_movements_DASH_in_absolute_coordinates(self):
        func = lambda specimens: self._batch_measure(specimens, absolute_coordinates=True)
        select_specimens(self.core, func, with_rois=True)



    def averaged_vectormap_DASH_interactive_plot(self):
        select_specimens(self.core, lambda specimens: self.core.adm_subprocess(specimens, '--tk_waiting_window --average -A vectormap'), with_movements=True)


    
    def averaged_vectormap_DASH_rotating_video(self):
        select_specimens(self.core, lambda specimens: self.core.adm_subprocess(specimens, '--tk_waiting_window --average -A vectormap_video'), with_movements=True) 


    def averaged_vectormap_DASH_rotating_video_multiprocessing(self):
        
        def run_workers(specimens):
            if len(specimens) > 0:
                N_workers = os.cpu_count()
                for i_worker in range(N_workers):
                    if i_worker != 0:
                        additional = '--dont-show'
                    else:
                        additional = ''
                    self.core.adm_subprocess(specimens, '--tk_waiting_window --worker-info {} {} --average -A vectormap_video'.format(i_worker, N_workers)) 


        select_specimens(self.core, run_workers, with_movements=True) 
        

    def averaged_vectormap_DASH_rotating_video_DASH_set_title(self):
        ask_string('Set title', 'Give video title', lambda title: select_specimens(self.core, lambda specimens: self.core.adm_subprocess(specimens, '--tk_waiting_window --average --short-name {} -A vectormap_video'.format(title)), with_movements=True)) 
 

    def averaged_vectormap_DASH_export(self):
        def target(specimens):
            fn = tk.filedialog.asksaveasfilename(initialfile='average', defaultextension='.npy')
            if fn:
                self.core.adm_subprocess(specimens, f'--average -A export_vectormap --output {fn}')
        select_specimens(self.core, target)
        


    def compare_vectormaps(self):
        popup(self.tk_root, CompareVectormaps, args=[self.core],
                title='Vectormap comparison')
       

    def comparision_to_optic_flow_DASH_video(self): 
        select_specimens(self.core, lambda specimens: self.core.adm_subprocess(specimens, '--tk_waiting_window --average -A flow_analysis_pitch'), with_movements=True) 
        
    
    
    def export_LR_displacement_CSV(self, strong_weak_division=False):
        '''
        Grouped to left right
        '''
        def callback(specimens):
            group_name = ask_string('Group name', 'Name the selected group of specimens', self.tk_root)
            analysers = self.core.get_manalysers(specimens)
            left_right_displacements(analysers, group_name,
                    strong_weak_division=strong_weak_division) 

        select_specimens(self.core, callback, with_movements=True) 
    

    def export_LR_displacement_CSV_DASH_strong_weak_eye_division(self):
        '''
        Grouped to strong vs weak eye.
        '''
        self.export_LR_displacement_CSV(strong_weak_division=True)


    def save_kinematics_analysis_CSV(self, fit_to_mean=False):

        def callback(specimens):
            
            fn = tk.filedialog.asksaveasfilename(title='Save kinematics analysis', initialfile='kinematics.csv')
            
            if fn:
                analysers = self.core.get_manalysers(specimens)
                kinematics.save_sigmoidal_fit_CSV(analysers, fn, fit_to_mean=fit_to_mean)
 
        select_specimens(self.core, callback, with_movements=True) 
    

    def save_kinematics_analysis_CSV_DASH_fit_to_mean(self, ):
        self.save_kinematics_analysis_CSV(fit_to_mean=True)


    def save_sinesweep_analysis_CSV(self):
        def callback(specimens):
            analysers = self.core.get_manalysers(specimens)
            sinesweep.save_sinesweep_analysis_CSV(analysers)
 

        select_specimens(self.core, callback, with_movements=True) 


    def response_magnitude_stats(self):
        def callback(grouped_manalysers):
            response_magnitudes(grouped_manalysers)

        select_specimen_groups(self.core, callback)


    def LR_stasts(self):

        fns = filedialog.askopenfilenames(initialdir=ANALYSES_SAVEDIR)
        
        if fns:
            lrfiles_summarise(fns)
    
    def LR_kinematics(self):

        fns = filedialog.askopenfilenames(initialdir=ANALYSES_SAVEDIR)
        
        if fns:
            lrfiles_summarise(fns, point_type='kinematics')


    def export_wizard(self):
        top, wizard = popup(self.tk_root, ExportWizard,
                            args=[self.core], title='Export Wizard')
        
   

class OtherCommands(ModifiedMenuMaker):
    '''
    All kinds of various commands and tools.
    '''
    
    def _force_order(self):
        return ['manage_specimen_groups',
                'link_ERG_data_from_labbook',
                '.',
                'change_Analyser_DASH_object',
                '.',
                'about']


    def manage_specimen_groups(self):
        '''
        Fixme: This is a little hacked together.
        '''
        
        def _preedit():
            select_specimens(self.core, _postedit)

        def _postedit(specimens):
            self.dm.im2.set_data(specimens)
            self.dm.im2.postchange_callback(self.dm.im2.data)

        def onsave():
            self.groups.groups = self.dm.im1.data
            self.groups.save_groups()


        def oncancel():
            top.destroy()

        self.groups = SpecimenGroups() 
        
        top = tk.Toplevel(self.tk_root)

        self.dm = ListManager(top, start_data=self.groups.groups,
                save_callback=onsave, cancel_callback=oncancel)
        tk.Button(self.dm.im2.buttons, text='Select specimens', command=_preedit).grid(row=3, column=1)
        self.dm.grid(row=1, column=1, sticky='NSWE')
        top.rowconfigure(1, weight=1)
        top.columnconfigure(1, weight=1)

        top.mainloop()

    
    def link_ERG_data_from_labbook(self):
        select_specimens(self.core, linked_data.link_erg_labbook, command_args=[lambda: filedialog.askopenfilename(title='Select ERG'), lambda: filedialog.askdirectory(title='Select data folder')], return_manalysers=True )
 


    def change_Analyser_DASH_object(self):

        popup_tickselect(self.tk_root,
                [c.__name__ for c in self.core.analyser_classes],
                lambda selections: self.core.set_analyser_class(selections[0]),
                ticked=[self.core.analyser_class],
                single_select=True)


    def about(self):
        message = 'Gonio analysis'
        message += "\nVersion {}".format(gonioanalysis.__version__)
        message += '\n\nGPL-3.0 License'
        tk.messagebox.showinfo(title='About', message=message)




import os
import csv

import numpy as np
import scipy.optimize

import matplotlib.pyplot as plt

def _drop_trace(displacement):
    # Improve fitting in conditions that motion noise is comparable to the
    # motion response: Remove the point zero and drop down by the noise level
    # at the beginning (the first three frames)
    displacement[0] = displacement[1]
    displacement = displacement - np.mean(displacement[:3])
    return displacement



def _logistic_function(t, k, L, t0):
    '''
    The sigmoidal function.

    t : float or np.ndarray
        Independent variable, time
    k : float
        Steepness of the curve
    L : float
        Curve's maximum value
    t0: : float
        Timepoint of the mid value

    See https://en.wikipedia.org/wiki/Logistic_function
    '''
    return L / (1+np.exp(-k*(t-t0)))



def mean_max_response(manalyser, image_folder, maxmethod='max'):
    '''
    Averages over repetitions, and returns the maximum of the
    mean trace.

    manalyser
    image_folder
    maxmethod : string
        Method how the determine the maximum displacement
        'max': Take the furthest point
        'mean_latterhalf': final 50% mean discplacement
        'final' : Use the final value
    '''
    
    displacements = manalyser.get_displacements_from_folder(image_folder)
    
    mean = np.mean(displacements, axis=0)
    if maxmethod == 'max':
        return np.max(mean)
    elif maxmethod == 'mean_latterhalf':
        return np.mean(mean[int(len(mean)/2):])
    elif maxmethod == 'final':
        return mean[-1]
    else:
        raise ValueError



def magstd_over_repeats(manalyser, image_folder, maxmethod='max'):
    '''
    Standard deviation in responses
    (std of max displacement of each repeat)
    
    maxmethod : string
        See mean_max_response
    '''
    displacements = manalyser.get_displacements_from_folder(image_folder)
    
    if maxmethod == 'max':
        displacements = np.max(displacements, axis=1)
    elif maxmethod == 'mean_latterhalf':
        displacements = np.mean([d[int(len(d)/2):] for d in displacements], axis=1)

    return np.std(displacements)


def mean_topspeed(manalyser, image_folder):
    '''
    Returns the top speed of the mean response.
    '''
    mean = np.mean(manalyser.get_displacements_from_folder(image_folder), axis=0)
    
    # FIXME: Replace the loop with numpy routines
    top = 0
    for i in range(len(mean)-1):
        top = max(top, mean[i+1] - mean[i])
    
    return top


def _simple_latencies(displacements, fs, threshold=0.1):
    '''
    Interpolate and take when goes over 1/10th
    ''' 
    latencies = []

    timepoints = np.linspace(0, len(displacements[0])/fs, len(displacements[0]))
    newx = np.linspace(0, len(displacements[0])/fs, 200)

    for displacement in displacements:
        y = np.interp(newx, timepoints, displacement)
        
        index = np.where(y>(np.max(y)*threshold))[0][0]
        
        latencies.append(newx[index])
        
    return latencies



def latency(manalyser, image_folder, threshold=0.05, method='sigmoidal',
            fit_to_mean=False):
    '''
    Response latency ie. when the response exceedes by default 5% of
    its maximum value.
    
    Arguments
    ---------
    threshold : float
        Between 0 and 1
    method : string
        Either "sigmoidal" (uses the sigmoidal fit) or
        "simple" (uses the data directly).
    fit_to_mean : bool
        If method=='sigmoidal', performs the fit on the image folder's mean
        and not on the responses directly.
    
    Returns
    -------
    latency : sequency 
        The time durations in seconds that it takes for the responses
        to reach (by default) 5% of its maximum value (length of repeats).
    '''    
    fs = manalyser.get_imaging_frequency(image_folder)
    
    if method == 'simple':
        # Take the mean response of the image_folder's data
        displacements = manalyser.get_displacements_from_folder(image_folder)
        trace = np.mean(displacements, axis=0)
        trace = _drop_trace(trace)
    elif method == 'sigmoidal':
        # Make a sigmoidal fit and use the sigmoidal curve
        params = sigmoidal_fit(manalyser, image_folder, fit_to_mean=fit_to_mean)
        N = len(manalyser.get_displacements_from_folder(image_folder)[0])
        time = np.linspace(0, N/fs, N)
        trace = _logistic_function(time,
                np.mean(params[1]),
                np.mean(params[0]),
                np.mean(params[2]))
    else:
        raise ValueError("method has to be 'sigmoidal' or 'simple', not {}".format(
            method))

    # Check when climbs over the threshold
    latency = _simple_latencies([trace], fs, threshold)
    return latency



def _sigmoidal_fit(displacements, fs, debug=False):
    amplitudes = []
    speeds = []
    latencies = []

    if debug:
        fig, ax = plt.subplots()

    timepoints = np.linspace(0, len(displacements[0])/fs, len(displacements[0]))

    for i_repeat, displacement in enumerate(displacements):
        
        # Initial guesses for k,L,t0
        est_L = displacement[-1]
        if est_L > 0:
            est_t0 = (np.argmax(displacement)/fs)/2
        else:
            est_t0 = (np.argmin(displacement)/fs)/2
        est_k = abs(est_L/est_t0)
        
        print('est L={} t0={} k={}'.format(est_L, est_t0, est_k))

        try:
            popt, pcov = scipy.optimize.curve_fit(_logistic_function, timepoints, displacement,
                    p0=[est_k, est_L, est_t0])
        except RuntimeError:
            # Runtime Error occurs when curve fitting takes over maxfev iterations
            # Usually then we have nonsigmoidal data (no response)
            continue
       
       
        if debug:
            ax.plot(timepoints, displacement, '-')
            ax.plot(timepoints, _logistic_function(timepoints, *popt),
                    '--', label='fit rep {}'.format(i_repeat))
            
            plt.show(block=False)
            plt.pause(.1)
            if not input('good?')[0].lower() == 'y':
                plt.close(fig)
                return None
            plt.close(fig)

        speeds.append(popt[0])
        amplitudes.append(popt[1])
        latencies.append(popt[2])
 

    return amplitudes, speeds, latencies



def sigmoidal_fit(manalyser, image_folder, figure_savefn=None, fit_to_mean=False):
    '''

    Assuming sigmoidal (logistic function) response.
    
    Arguments
    ---------
    manalyser : object
    image_folder : string
    figure_savefn : string
        If given, saves a figure of the fit

    Returns
    -------
    amplitudes, speeds, halfrise_times : list or None
        All Nones if image_folder has not movements
    fit_to_mean : bool
        If True, calculate the mean for the image folder and perform the
        fit on the mean.
    '''

    if figure_savefn is not None:
        fig, ax = plt.subplots()


    amplitudes = []
    speeds = []
    halfrise_times = []
    
    pcovs = []
    
    displacements = manalyser.get_displacements_from_folder(image_folder)
    if not displacements:
        # Probably movements not measured
        return None, None, None

    if fit_to_mean:
        displacements = [np.mean(displacements, axis=0)]

    fs = manalyser.get_imaging_frequency(image_folder)

    timepoints = np.linspace(0, len(displacements[0])/fs, len(displacements[0]))
    
    for i_repeat, displacement in enumerate(displacements):
        _drop_trace(displacement)
        amplitude, speed, halfrise_time = _sigmoidal_fit([displacement], fs)
        
        if not amplitude or not speed or not halfrise_time:
            print(f'Fit failed for repeat={i_repeat}. Data likely not sigmoidal.')
            continue


        speeds.append(speed[0])
        amplitudes.append(amplitude[0])
        halfrise_times.append(halfrise_time[0])

        if figure_savefn is not None:
            ax.plot(timepoints, displacement, '-')
            ax.plot(
                    timepoints,
                    _logistic_function(timepoints, speed[0], amplitude[0], halfrise_time[0]),
                    '--', label='fit rep {}'.format(i_repeat)
                    )
    

    if figure_savefn:
        fig.savefig(figure_savefn)

    return amplitudes, speeds, halfrise_times



def save_sigmoidal_fit_CSV(analysers, savefn, save_fits=True, with_extra=True,
        microns=True, fit_to_mean=False):
    '''
    Takes in analysers, performs sigmoidal_fit for each and all image_folders.
    Then saves the results as a CSV file, and by default fit images as well.

    analysers : list of objects
        List of analyser objects
    savefn : string
        Filename.
    save_fits : bool
        Save png images of the fits.
    fit_to_mean : bool
        If True, calculate the mean for the image folder and perform the
        fit on the mean.
    '''
    
    with open(savefn, 'w') as fp:
        writer = csv.writer(fp)
    
        writer.writerow(['#'])
        writer.writerow(['# Kinematics by sigmoidal fit', 'L / (1+np.exp(-k*(t-t0)))'])
        writer.writerow(['# t0 latency (s)', 'k rise speed (pixels/s)', 'L response amplitude (pixels)'])
        writer.writerow(['#'])

        if not microns:
            y_units = 'pixels'
        else:
            y_units = 'µm'

        header = ['Name', 'Image folder',
                'Mean L ({})', 'STD L ({})',
                'Mean k ({}/s)', 'STD k ({}/s)',
                'Mean t0 (s)', 'STD t0 (s)', 'Fit image']

        if with_extra:
            header.extend( ['Mean max amplitude ({})',
                'Mean final amplitude ({})' ,
                'Mean latency simple (s)',
                'Mean latency sigfit (s)',
                'Top speed ({}/s)'] )

        writer.writerow([text.format(y_units) for text in header])
        
        i_fit = 0

        for analyser in analysers:
            try: 
                for image_folder in analyser.list_imagefolders():
                    if save_fits:
                        dirname = os.path.dirname(savefn)
                        folder = os.path.basename(dirname)
                        figure_savefn = os.path.join(dirname, folder+'_fits', 'fit_{0:07d}.png'.format(i_fit))
                        
                        if i_fit == 0:
                            os.makedirs(os.path.dirname(figure_savefn), exist_ok=True)

                    else:
                        figure_savefn = False
                    
                    # Fixme
                    if 'logsweep' in image_folder:
                        continue

                    if not analyser.folder_has_movements(image_folder):
                        continue

                    amplitudes, speeds, halfrise_times = sigmoidal_fit(analyser, image_folder,
                            figure_savefn=figure_savefn, fit_to_mean=fit_to_mean)
                    
                    if microns:
                        scaler = analyser.get_pixel_size(image_folder)
                    else:
                        scaler = 1
                    
                    if figure_savefn:
                        figure_savefn = os.path.basename(figure_savefn)

                    row = [analyser.folder, image_folder,
                            np.mean(amplitudes)*scaler, np.std(amplitudes)*scaler,
                            np.mean(speeds)*scaler, np.std(speeds)*scaler,
                            np.mean(halfrise_times), np.std(halfrise_times),
                            figure_savefn]

                    if with_extra:
                        max_amplitude = scaler * mean_max_response(analyser, image_folder, maxmethod='max')
                        end_amplitude = scaler * mean_max_response(analyser, image_folder, maxmethod='final')
                        latency_value1 = np.mean(latency(analyser, image_folder, fit_to_mean=fit_to_mean, method='simple'))
                        latency_value2 = np.mean(latency(analyser, image_folder, fit_to_mean=fit_to_mean, method='sigmoidal'))
                        
                        top_speed = scaler * mean_topspeed(analyser, image_folder)

                        row.extend([max_amplitude, end_amplitude, latency_value1, latency_value2, top_speed])

                    writer.writerow(row)
            except Exception as e:
                print(e)

            i_fit += 1

    return None




'''
Whole gonioanalysis wide settings here.

Attributes
-----------

DEFAULT_FILENAME
'''

import os
import json

from gonioanalysis.directories import GONIODIR


DEFAULT_SAVENAME = "gonioanalysis-settings.json"


def _load(fn):
    fullfn = os.path.join(GONIODIR, fn)
    if os.path.isfile(fullfn):
        with open(fullfn, 'r') as fp:
            return json.load(fp)
    else:
        return {}

def _save(data, fn):
    with open(os.path.join(GONIODIR, fn), 'w') as fp:
        json.dump(data, fp)


def set(key, value, fn=DEFAULT_SAVENAME):
    '''
    Set and save a setting.
    '''
    settings = _load(fn)

    settings[key] = value

    _save(settings, fn)


def get(key, default=None, fn=DEFAULT_SAVENAME):
    '''
    Get a setting.
    Specidy default for default value like in Python's standard get.
    '''
    settings = _load(fn)

    return settings.get(key, default)




'''
Analysing sinusoidal sweep data.

Currently supports Gonio Imsoft flash_type's 
'''

import os
import csv
import math

import numpy as np
import scipy.stats
import scipy.signal
import scipy.interpolate
import matplotlib.pyplot as plt


from gonioanalysis.directories import ANALYSES_SAVEDIR

def _get_stimulus(flash_type, t, fs):
    '''
    
    flash_type : strings
        Flash_type used in Gonio Imsoft
    t : float
        Duration of the stimulus
    fs : float
        Sampling rate of the stimulus
    '''

    if ',' in flash_type:
        flash_type, f0, f1 = flash_type.split(',')
        f0 = float(f0)
        f1 = float(f1)
    else:
        f0 = 0.5
        f1 = 100

    
    timepoints = np.linspace(1/fs, t, int(t*fs))

    stimulus_frequency = f0 * (f1/f0) ** (timepoints/t)
    stimulus_amplitude = scipy.signal.chirp(timepoints, f0=f0, t1=t, f1=f1,
            phi=-90, method='logarithmic')
    if flash_type == 'squarelogsweep':
        stimulus_amplitude[stimulus_amplitude>=0] = 1
        stimulus_amplitude[stimulus_amplitude<0] = -1
    elif flash_type == '3steplogsweep':
        cstep = np.sin(np.pi/4)
        stimulus_amplitude[np.abs(stimulus_amplitude) <= cstep] = 0
        stimulus_amplitude[stimulus_amplitude > cstep] = 1
        stimulus_amplitude[stimulus_amplitude < -cstep] = -1
    else:
        pass

    stimulus_amplitude = (stimulus_amplitude+1)/2

    return timepoints, stimulus_frequency, stimulus_amplitude


from numpy.fft import fftshift

def _find_zeroindices(stimulus):
    '''
    Poor mans iterative zero point finding.
    '''
    zero_indices = []

    previous_dir = 0

    for i in range(len(stimulus)):
        dirr = np.sign(stimulus[i] - stimulus[i-1])
        
        if dirr != previous_dir:
            zero_indices.append(i)

            previous_dir = dirr
        else:
            pass
    return zero_indices


def _sham_frequency_response(stimulus_timepoints, stimulus_frequencies, stimulus, response,
        interpolate=True):
    '''
    Calculates frequency response to sinusoidal sweep signal.
    
    interpolate : bool
        Interpolate to stimulus_frequencies
    '''
    #fn = '/home/joni/.gonioanalysis/final_results/sineweep_analysis/wtb_sinusoidal_07_right_sinelogsweep.csv'
    #fn = '/home/joni/.gonioanalysis/final_results/sineweep_analysis/wtb_sinusoidal_07_right_squarewave_cam200Hz.csv'
    #data = np.loadtxt(fn,
    #        skiprows=2, delimiter=',').T

    #print(data)

    #stimulus_timepoints = data[1][:-5]
    #stimulus_frequencies = data[2][:-5]
    #stimulus = data[3][:-5]
    #response = data[4][:-5]

    fs = 1 / (stimulus_timepoints[1]-stimulus_timepoints[0])

    #b, a = scipy.signal.butter(1, 0.5, 'high', fs=fs)
    #response = scipy.signal.filtfilt(b, a, response)

    #calculate_frequency_response(stimulus_timepoints, stimulus)


    cut_indices = _find_zeroindices(stimulus)[1::2]

    #for indx in cut_indices:
    #    plt.plot(2*[stimulus_timepoints[indx]], [0, 1], '--', color='black')

    #plt.plot(stimulus_timepoints, response, color='red')
    #plt.plot(stimulus_timepoints, stimulus, color='blue')
    #plt.show()

    freqs = []
    resps = []

    for i1, i2 in zip(cut_indices[0:-1], cut_indices[1:]):
        #plt.plot(stimulus_timepoints[i1:i2], stimulus[i1:i2], color='blue')
        #plt.plot(stimulus_timepoints[i1:i2], response[i1:i2], color='red')
        #plt.plot(stimulus_timepoints, response2, color='orange')
        #plt.show()
        chunk = response[i1:i2]
        resp = max(chunk) - min(chunk)
        
        freq = np.mean(stimulus_frequencies[i1:i2])

        resps.append(resp)
        freqs.append(freq)
    
    #plt.plot(freqs, resps)
    #plt.xscale('log')
    #plt.show()

    if interpolate:
        f = scipy.interpolate.interp1d(freqs, resps, fill_value='extrapolate', bounds_error=False)
        resps = f(stimulus_frequencies)
        freqs = stimulus_frequencies
    
    return freqs, resps



def save_sinesweep_analysis_CSV(analysers, debug=False):
    '''
    Save X and Y components of the movement in a csv file
    for the selected manalysers.

    Columns
        i_camframe time stimulus_fs mean_response response_rep1, response_rep2...
    '''

    stimuli = {}

    savedir = os.path.join(ANALYSES_SAVEDIR, 'sineweep_analysis')
    os.makedirs(savedir, exist_ok=True)
   
    final_cols = {}

    if debug:
        fig, ax = plt.subplots()

    for analyser in analysers:
        for eye in analyser.eyes:

            for image_folder in analyser.list_imagefolders():
                
                imagefolder_key = image_folder[3:]

                try:
                    N_repeats = len(analyser.movements[eye][imagefolder_key])
                except KeyError:
                    print('No data for {}'.format(analyser))
                    continue

                # Get imaging parameters
                im_fs = analyser.get_imaging_frequency(image_folder)

                im_params = analyser.get_imaging_parameters(image_folder)
                flash_type = im_params.get('flash_type', '')
               
                if flash_type.split(',')[0] not in ['squarelogsweep', 'sinelogsweep', '3steplogsweep']:
                    print('Unkown flash_type {}, skipping'.format(flash_type))
                    continue


                # Get movement data
                pixel_size = analyser.get_pixel_size(image_folder)
                print('N_repeats {}'.format(N_repeats))
                Xs = [np.array(analyser.movements[eye][imagefolder_key][i_repeat]['x'])*pixel_size for i_repeat in range(N_repeats)]
                Ys = [np.array(analyser.movements[eye][imagefolder_key][i_repeat]['y'])*pixel_size for i_repeat in range(N_repeats)]

                #mean_displacement = np.mean([np.sqrt(np.array(X)**2+np.array(Y)**2) for X, Y in zip(Xs, Ys)], axis=0)

                meanX = np.mean(Xs, axis=0)
                meanY = np.mean(Ys, axis=0)
                meanX -= meanX[0]
                meanY -= meanY[0]

                mean_displacement = np.sqrt(meanX**2+meanY**2)
                
                #slope, intercept, r_value, p_value, std_err = scipy.stats.linregress(meanX, meanY)
                
                # Rotate fit along xaxis
                #rot = -math.atan(slope)
                #c = math.cos(rot)
                #s = math.sin(rot)
                #pc1 = [c*x - s*y for x,y in zip(meanX, meanY)]
                #pc2 = [s*x + c*y for x,y in zip(meanX, meanY)]
                
                # Stimulus frequencies
                #stim_fs = 1000
                

                timepoints, stimulus_frequency, stimulus_amplitude = _get_stimulus(flash_type, len(Xs[0])/im_fs, im_fs)
                
                if flash_type not in stimuli:
                    # FIXME 1kHz output in Imsoft but information not saved
                    stim_fs = 1000
                    dense = _get_stimulus(flash_type, len(Xs[0])/im_fs, stim_fs)
                    
                    stimuli[flash_type] = dense

                # "Frequency response"
                fr_freqs, fr = _sham_frequency_response(timepoints, stimulus_frequency,
                        stimulus_amplitude, mean_displacement, interpolate=True)

                # Save csv

                fn = '{}_{}_{}.csv'.format(analyser.folder, eye, im_params['suffix'])
                
                if debug:
                    ax.clear()
                    ax.scatter(meanX, meanY)
                    ax.plot(meanX, intercept + slope*meanX)
                    fig.savefig(os.path.join(savedir, fn.replace('.csv', '.png')))

                if final_cols.get(flash_type, None) is None:
                    final_cols[flash_type] = {'time (s)': [], 'f_stimulus (Hz)': [],
                            'displacement': [], 'frequency_response': [], 'specimen_name': []}
                

                final_cols[flash_type]['time (s)'].append(timepoints)
                final_cols[flash_type]['f_stimulus (Hz)'].append(stimulus_frequency)
                final_cols[flash_type]['displacement'].append(mean_displacement)
                final_cols[flash_type]['specimen_name'].append(analyser.folder)
                final_cols[flash_type]['frequency_response'].append(fr)
                

                with open(os.path.join(savedir, fn), 'w') as fp:
                    writer = csv.writer(fp)
                    
                    writer.writerow(['Displacement = sqrt(X_mean**2+Y_mean**2)'])
                    
                    row = ['i_camframe',
                            'time (s)',
                            'f_stimulus (Hz)',
                            'instantaneous stimulus amplitude (0-1)',
                            #'pc1 (µm)',
                            #'displacement',
                            'displacement2 (µm)',
                            #'pc2 (µm)'
                            #'X_mean (µm)',
                            #'Y_mean (µm)',
                            #'Displacement (µm)'
                            'Frequency response (µm)']

                    for i_repeat, (x, y) in enumerate(zip(Xs, Ys)):
                        row.append('X rep_{}'.format(i_repeat))
                        row.append('Y rep_{}'.format(i_repeat))

                    writer.writerow(row)

                    for i in range(len(Xs[0])-10):
                        
                        row = []
                        row.append(i)
                        row.append(timepoints[i])
                        row.append(stimulus_frequency[i])
                        row.append(stimulus_amplitude[i])
                        #row.append(pc1[i])
                        #row.append(pc2[i])
                        #row.append(meanX[i])
                        #row.append(meanY[i])
                        #row.append(math.sqrt(meanX[i]**2+meanY[i]**2))
                        row.append(mean_displacement[i])
                        row.append(fr[i])

                        for x, y in zip(Xs, Ys):
                            row.append(x[i])
                            row.append(y[i])

                        writer.writerow(row)
                    
    
    for flash_type in final_cols:
        with open(os.path.join(savedir, 'mean_'+flash_type+'.csv'), 'w') as fp:
            writer = csv.writer(fp)
            
            row = []
            row.append('i_camframe')
            row.append('time (s)')
            row.append('f_stimulus (Hz)')
            
            N = len(final_cols[flash_type]['displacement'])
            
            for k in range(N):
                row.append('{} response (µm)'.format(final_cols[flash_type]['specimen_name'][k]))
            
            row.append('mean response (µm)')

            for k in range(N):
                row.append('frequency response (µm)')


            row.append('mean frequency response (µm)')

            writer.writerow(row)

            for i in range(len(final_cols[flash_type]['time (s)'][0])-10):

                row = []
                row.append(i)
                row.append(final_cols[flash_type]['time (s)'][0][i])
                row.append(final_cols[flash_type]['f_stimulus (Hz)'][0][i])
                

                displacements = []
                for j in range(N):
                    row.append(final_cols[flash_type]['displacement'][j][i])
                    displacements.append(row[-1])

                row.append(np.mean(displacements))
                

                fr = []
                for j in range(N):
                    row.append(final_cols[flash_type]['frequency_response'][j][i])
                    fr.append(row[-1])
                row.append(np.mean(fr))

                writer.writerow(row)
    
    for stimulus in stimuli:

        with open(os.path.join(savedir, 'stimulus_{}.csv'.format(stimulus)), 'w') as fp:
            writer = csv.writer(fp)
            writer.writerow(['Time (ms)', 'Frequency (Hz)', 'Amplitude (V)'])
            for i in range(len(stimuli[stimulus][0])):
                writer.writerow([data[i] for data in stimuli[stimulus]])
                





from math import sin, cos, tan, sqrt, radians, pi

import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits import mplot3d

from gonioanalysis.coordinates import camera2Fly, optimal_sampling


def test():
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    
    verticals = [1, 30, 45, 60, 90]
    colors = ['red', 'orange', 'yellow', 'green', 'blue']
    for vertical, color in zip(verticals, colors):
        horizontals = np.linspace(-90, 90)

        for horizontal in horizontals:
            x,y,z = camera2Fly(horizontal, vertical)
            ax.scatter(x,y,z, color=color)
        
    plt.show()

def plot_optimal(points):


    print(len(points))

    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')

    ax.scatter(*list(zip(*points)))
    
    plt.show()



if __name__ == "__main__":

    steps = (10, 10)


    # For Gabor
    #horizontals = np.arange(-90,40+0.01, steps[0])
    #verticals = np.arange(-20,180+0.01, steps[1])
    
    # For vector map
    horizontals = np.arange(-50,40+0.01, steps[0])
    verticals = np.arange(-20,180+0.01, steps[1])
    
    # Full sphere just for illustration
    #horizontals = np.arange(-90,90+0.01, steps[0])
    #verticals = np.arange(0,360+0.01, steps[1])
 
    points = optimal_sampling(horizontals, verticals, steps)
    plot_optimal(points)



'''
Common helper functions likely needed in many different plots.
'''

import os
import math
import copy
import multiprocessing
import datetime

import numpy as np
from scipy.spatial import cKDTree as KDTree
import matplotlib.pyplot as plt
import matplotlib.animation
import matplotlib.colors
from matplotlib.patches import FancyArrowPatch, CirclePolygon
from mpl_toolkits.mplot3d import proj3d, art3d
from matplotlib import cm
 
from gonioanalysis.coordinates import (
        nearest_neighbour,
        get_rotation_matrix,
        rotate_points
        )
from gonioanalysis.directories import ANALYSES_SAVEDIR
from gonioanalysis.version import used_scipy_version


CURRENT_ARROW_LENGTH = 1

VECTORMAP_PULSATION_PARAMETERS = {'step_size': 0.02, 'low_val': 0.33, 'high_val': 1}



# Taken from drosoeyes.blend
RHABDOMERE_LOCATIONS = [(-1.6881, 1.0273), (-1.8046, -0.9934),
        (-1.7111, -2.9717), (-0.0025, -1.9261), (1.6690, -0.9493),
        (1.6567, 0.9762), (0.0045, -0.0113)]
RHABDOMERE_DIAMETERS = [1.8627,1.8627,1.8627,1.8627,1.8627,1.8627, 1.5743]
RHABDOMERE_R3R6_ROTATION = math.radians(-49.7)



class Arrow3D(FancyArrowPatch):
    def __init__(self, x0, y0, z0, x1, y1, z1, *args, **kwargs):
        FancyArrowPatch.__init__(self, (0,0), (0,0), *args, **kwargs)
        self._verts3d = (x0, x1), (y0, y1), (z0, z1)

    def draw(self, renderer):
        xs3d, ys3d, zs3d = self._verts3d
        xs, ys, zs = proj3d.proj_transform(xs3d, ys3d, zs3d, self.axes.M)
        self.set_positions((xs[0],ys[0]),(xs[1],ys[1]))
        FancyArrowPatch.draw(self, renderer)

    def do_3d_projection(self, renderer=None):
        '''Fix for newer matplotlib versions.
        See https://github.com/matplotlib/matplotlib/issues/21688
        '''
        xs3d, ys3d, zs3d = self._verts3d
        xs, ys, zs = proj3d.proj_transform(xs3d, ys3d, zs3d, self.axes.M)
        self.set_positions((xs[0],ys[0]),(xs[1],ys[1]))

        return np.min(zs)


def add_line(ax, x0, y0, z0, x1, y1, z1, camerapos=None, **kwargs):
    '''
    Add a centered 3D line plot
    '''
    if camerapos and is_behind_sphere(*camerapos, (x0,y0,z0)):
        return None

    ax.plot([x0-x1/2, x0, x0+x1/2], [y0-y1/2, y0, y0+y1/2], [z0-z1/2, z0, z0+z1/2], **kwargs)


def add_rhabdomeres(ax, x0, y0, z0, x1, y1, z1, mirror_lr=False, mirror_bf=False,
        scale=0.015, camerapos=None, **kwargs):
    '''
    Add rhabdomeres R1-R7/8 patches to a 3D plot.
    
    ax : object
        Matplotlib Axes object to add attach the patches onto
    x0,y0,z0 : float
        Coordinates of the rhabdomeres center point (R6/R7)
    x1,y1,z1 : float
        Vector pointing the direction of R6R3 line
    mirror_lr, mirror_bf : True, False (or auto for bf)
        Wheter to mirror this rhabdomere or not (left/right, back/front)
    scale : float
        Scale of the rhabdemeres
    **kwargs : dict
        To matplotlib.patches.CirclePolygon

    Returns a list of matplotlib 3d patches
    '''
    
    if camerapos and is_behind_sphere(*camerapos, (x0,y0,z0)):
        return None


    #v = np.asarray([x0,y0,z0])
    #uv = v / np.linalg.norm(v)
    
    try:
        phi = math.asin(z0)
    except:
        phi = math.pi / 2

    try:
        theta = math.atan(x0/y0)
    except:
        theta = math.pi / 2
    if y0 < 0:
        theta = theta + math.pi


    # Calculate rhabdomere rotation to match x1,y1,z1 by transforming
    # point [1,0,0] to the x0,y0,z0 point and calculating angle of
    # transformed ux and x1,y1,z1
    ux = np.array([1,0,0])
    ux = get_rotation_matrix('x', phi) @ ux
    ux = get_rotation_matrix('z', -theta) @ ux
    rot = np.arccos(np.inner(ux, [x1,y1,z1])/(np.linalg.norm(ux) * np.linalg.norm([x1,y1,z1])))
    
    if z1 < 0:
        rot = -rot

    patches = []
    
    if mirror_bf == 'auto' and z0 > 0:
        mirror_bf = True
    elif mirror_bf is not True:
        mirror_bf = False

    for diameter, location in zip(RHABDOMERE_DIAMETERS, RHABDOMERE_LOCATIONS):
        
        patch = CirclePolygon((location[0]*scale, location[1]*scale), diameter/2*scale,
                **kwargs)
        patches.append(patch)
        ax.add_patch(patch)

        art3d.patch_2d_to_3d(patch)
        
        #if mirror_lr:
        #    patch._segment3d = [get_rotation_matrix('y', math.pi) @ p for p in patch._segment3d] 

        # Rotate according to the vector (x1,y1,z1)
        #   First z rotation to set initial rotation 
        patch._segment3d = [get_rotation_matrix('z', RHABDOMERE_R3R6_ROTATION) @ p for p in patch._segment3d] 
        
        #if not mirror_lr and not mirror_bf:
        #    pass

        if mirror_lr and not mirror_bf:
            patch._segment3d = [get_rotation_matrix('x', math.pi) @ p for p in patch._segment3d] 
        
        if not mirror_lr and mirror_bf:
            patch._segment3d = [get_rotation_matrix('x', math.pi) @ p for p in patch._segment3d] 
            #patch._segment3d = [get_rotation_matrix('z', math.pi) @ p for p in patch._segment3d] 

        #if not mirror_lr and mirror_bf:
        #    patch._segment3d = [get_rotation_matrix('x', math.pi) @ p for p in patch._segment3d] 
        #    patch._segment3d = [get_rotation_matrix('z', math.pi) @ p for p in patch._segment3d] 

        #if mirror_lr and mirror_bf:
        #    patch._segment3d = [get_rotation_matrix('z', math.pi) @ p for p in patch._segment3d] 
        


        patch._segment3d = [get_rotation_matrix('z', rot) @ p for p in patch._segment3d] 
        
        patch._segment3d = [get_rotation_matrix('x', math.pi/2) @ p for p in patch._segment3d] 
        

        patch._segment3d = [get_rotation_matrix('x', phi) @ p for p in patch._segment3d]
        patch._segment3d = [get_rotation_matrix('z', -theta) @ p for p in patch._segment3d]

        # Translate
        patch._segment3d = [(x+x0,y+y0,z+z0) for x,y,z in patch._segment3d]


    return patches


def make_animation_angles(step=0.5):
    '''
    Returns the matplotlib angles to rotate a 3D plot
    
    This really shouldnt be here...
    '''

    animation = []
    sidego = 30
    # go up, to dorsal
    for i in np.arange(-30,60,step):
        animation.append((i,90))
    #rotate azim
    for i in np.arange(90,90+sidego,step*2):
        animation.append((60,i))
    # go back super down, to ventral
    for i in np.arange(0,120,step):
        animation.append((60-i,90+sidego))
    # rotate -azim
    for i in np.arange(0,2*sidego,step*2): 
        animation.append((-60,90+sidego-i))
    # go up back to dorsal
    for i in np.arange(0,120, step):
        animation.append((-60+i,90-sidego))
    return animation



CURRENT_ARROW_DIRECTION = 1
def make_animation_timestep(step_size=0.075, low_val=0.6, high_val=1, rel_return_speed=3, twoway=False):
    '''
    step_size between 0 and 1.
    Once total displacement 1 has been reached go back to low_value
    '''
    # Update arrow length
    global CURRENT_ARROW_LENGTH
    global CURRENT_ARROW_DIRECTION

    step = step_size * 1.5
    if CURRENT_ARROW_DIRECTION < 0:
        step *= rel_return_speed

    CURRENT_ARROW_LENGTH += step

    if twoway:
        s = -1
    else:
        s = 1

    if CURRENT_ARROW_DIRECTION > 0:
        if CURRENT_ARROW_LENGTH > high_val*1.5:
            CURRENT_ARROW_LENGTH = low_val*1.5
            CURRENT_ARROW_DIRECTION = s * CURRENT_ARROW_DIRECTION
    else:
        if CURRENT_ARROW_LENGTH > high_val * 1.5: 
            CURRENT_ARROW_LENGTH = low_val*1.5
            CURRENT_ARROW_DIRECTION = s * CURRENT_ARROW_DIRECTION


def plot_2d_opticflow(ax, direction):
    
    x0, x1 = ax.get_xlim()
    y0, y1 = ax.get_ylim()

    width = abs(x1-x0)
    height = abs(y1-y0)

    if direction == 'side':
        arrows = [np.array((1.05,ik))*width for ik in np.arange(0.1,0.91,0.1)]
        for x, y in arrows:
            ax.arrow(x, y, -0.1*width, 0, width=0.01*width, color='darkviolet')
    else:
        x = []
        y = []
        for i in range(10):
            for j in range(10):
                x.append((0.5+i)*height/10)
                y.append((0.5+j)*width/10)
        ax.scatter(x,y, marker='x', color='darkviolet')


def is_behind_sphere(elev, azim, point):
    '''
    Calculates wheter a point seend by observer at (elev,azim) in spehrical
    coordinates is behind a sphere (radius == point) or not.
    
    NOTICE: Elev from horizontal plane (a non-ISO convention) and azim as in ISO
    '''

    cx = np.sin(np.radians(90-elev)) * np.cos(np.radians(azim))
    cy = np.sin(np.radians(90-elev)) * np.sin(np.radians(azim))
    cz = np.cos(np.radians(90-elev))
    
    vec_cam = (cx,cy,cz)
    vec_arr = point
    
    angle = np.arccos(np.inner(vec_cam, vec_arr)/(np.linalg.norm(vec_cam)*np.linalg.norm(vec_arr)))
    if angle > np.pi/2:
        return True
    else:
        return False


def plot_guidance(ax, camerapos=None, r=1,
        mutation_scale=6, hide_text=False):
    '''
    Plot help elements to point left,right,front,back
    '''
    arrows = []
    guidances = {'Right': ((r,0,0), (0.2,0,0)),
            'Left': ((-r,0,0),(-0.2,0,0)),
            '  Ventral': ((0,0,-r),(0,0,-0.3)),
            '  Dorsal': ((0,0,r),(0,0,0.3))}
       
    for name, (point, vector) in guidances.items():
        point = np.array(point)
        vector = np.array(vector)

        if is_behind_sphere(*camerapos, point):
            zorder = 1
        else:
            zorder = 8

        ar = Arrow3D(*point, *(point+vector), mutation_scale=mutation_scale,
                lw=0.2, color='black', zorder=zorder)
        ax.add_artist(ar)
        arrows.append(ar)
        
        if not hide_text:
            if name in ('Left', 'Right'):
                ha = 'center'
            else:
                ha = 'left'
            ax.text(*(point+vector/1.05), name, color='black',
                    fontsize='xx-large', va='bottom', ha=ha,
                    linespacing=1.5, zorder=zorder+1)

    return arrows


def plot_vrot_lines(ax, vrots, n_verts=16, camerapos=None):
    '''
    Plot vetical rotation lines

    Arguments
    ---------
    vrots:  list of floats
        Vertical rotations in degrees
    verts : int
        How many vertices per a half cirle. Higher values give
        smoother and more round results.
    '''
    
    horizontals = np.radians(np.linspace(-70, 70, n_verts))
    points = np.vstack( (np.sin(horizontals), np.cos(horizontals)) )
    points = np.vstack( (points, np.zeros(n_verts)) ).T
    
    points = points * 0.95

    print(points.shape)

    for vrot in vrots:
        pnts = rotate_points(points, 0, math.radians(vrot), 0)
        if -1 < vrot < 1:
            color = (0.2,0.2,0.2)
            style = '-'
        else:
            color = (0.2,0.2,0.2)
            style = '--'
        
        if camerapos:
            visible = [p for p in pnts if not is_behind_sphere(*camerapos, p)]
            
            if not visible:
                continue
            pnts = np.array(visible)

        ax.plot(pnts[:,0], pnts[:,1], pnts[:,2], style, lw=1, color=color)


def vector_plot(ax, points, vectors, color='black', mutation_scale=6, scale_length=1,
        i_pulsframe=None, guidance=False, camerapos=None, draw_sphere=True,
        vrot_lines=False,
        hide_axes=False, hide_text=False,
        **kwargs):
    '''
    Plot vectors on a 3D matplotlib Axes object as arrows.

    ax : object
        Matplotlib ax (axes) instance or Ax3d
    points : array_like
        Sequence of arrow starting/tail (x,y,z) points 
    vectors : array_like
        Arrow lengts and directions, sequence of (x,y,z)
    color : string
        Matplotlib valid color for the drawn arrows
    mutation_scale : float
        Size of the arrow head basically
    i_pulsframe : int
        Index of the pulsation frame, setting the arrow length. For animation.
    guidance : bool
        Add help elements to point left,right,front,back
    camerapos : tuple or None
        Values of (elev, axzim) to hide vectors behind the sphere. If none,
        use values from ax.elev and ax.azim.
    draw_sphere : bool
        If true draw a gray sphere
    hide_axes : bool
        Call set_axis_off
    hide_text : bool
        Omit from drawing any text
    kwargs : dict
        Passed to matplotlib FancyArrowPatch
    
    Returns
    -------
    arrow_artists : list
        All ArrowArtists added to the ax

    '''
    global CURRENT_ARROW_DIRECTION
    r = 0.9

    arrow_artists = []
    
    is_ax3d = hasattr(ax, 'add_arrow') and not hasattr(ax, 'add_artist')

    if hide_axes:
        ax.set_axis_off()
    
    if hide_text:
        if not is_ax3d:
            ax.axes.xaxis.set_ticklabels([])
            ax.axes.yaxis.set_ticklabels([]) 
            ax.axes.zaxis.set_ticklabels([]) 
        ax.set_xlabel('')
        ax.set_ylabel('')
        ax.set_zlabel('')
    else:
        ax.set_xlabel('x')
        ax.set_ylabel('y')
        ax.set_zlabel('z')



    #if not camerapos:
    #    camerapos = (ax.elev, ax.azim)
 
    if guidance:
        plot_guidance(ax, camerapos=camerapos, hide_text=hide_text)

    if draw_sphere:
        N = 75
        phi, theta = np.meshgrid(np.linspace(0, 2*np.pi, N), np.linspace(0, np.pi, N))
        X = r * np.sin(theta) * np.cos(phi)
        Y = r * np.sin(theta) * np.sin(phi)
        Z = r * np.cos(theta)
        ax.plot_surface(X, Y, Z, color='lightgray')

    
    if vrot_lines:
        plot_vrot_lines(ax, np.arange(-120, 120.1, 20), n_verts=16,
                camerapos=camerapos)


    if i_pulsframe:
        global CURRENT_ARROW_LENGTH
        scaler = CURRENT_ARROW_LENGTH
    else:
        scaler = 1.1

    scaler *= scale_length

    for point, vector in zip(points, vectors):

        if camerapos:
            vec_arr = point
    
            if is_behind_sphere(*camerapos, vec_arr):
                alpha = 0
            else:
                alpha = 1
                zorder = 10
        else:
            alpha = 1
            zorder = 10
        
        if CURRENT_ARROW_DIRECTION > 0 or i_pulsframe is None:
            A = point
            B = point+scaler*vector
        else:
            A = point
            B = point-scaler*vector
        
        if not is_ax3d:
            ar = Arrow3D(
                    *A, *B, arrowstyle="-|>", lw=1,
                    mutation_scale=mutation_scale, color=color,
                    alpha=alpha, zorder=10)
            ax.add_artist(ar)
            arrow_artists.append(ar)
        else:
            ax.add_arrow(*A, *B, color=color,
                         mutation_scale=mutation_scale)
    
    if not is_ax3d:
        ax.set_xlim(-1.1, 1.1)
        ax.set_ylim(-1.1,1.1)
        ax.set_zlim(-1.1, 1.1)
   
    return arrow_artists


def surface_plot(ax, points, values, cb=False, phi_points=None, theta_points=None,
        colormap='own', vlimits=None):
    '''
    3D surface plot of the error between the optic flow vectors and the actual
    eye-movements vector map.
    
    Points and values have to be in the same order.

    points
    values

    Arguments
    ---------
    ax : object
        Matplolib Axes object
    points : list
        List of x,y,z coordinates
    values : list of scalars
        List of scalar values at the given points, in the same order.
    colormap : string
        "own", "own-diverge" or any matplotlib colormap name
    vlimits : tuple or None
        If None, limits (0,1) are used
    '''

    if len(points) != len(values):
        raise ValueError('For sufrace_plot, points and values have to be same lenght (and in the same order)')
     
    
    if vlimits is None:
        vmin = 0
        vmax = 1
    else:
        vmin, vmax = vlimits
    

    # Points where the error is "evaluated" (actually interpolated)
    

    N = 100
    if phi_points is None:
        phi_points = np.linspace(0, 2*np.pi, N)
    
    phi, theta = np.meshgrid(phi_points, np.linspace(0, np.pi, N))
    X = np.sin(theta) * np.cos(phi)
    Y = np.sin(theta) * np.sin(phi)
    Z = np.cos(theta)
    

    def color_function(theta, phi):
        
        intp_dist = (2 * np.sin(np.radians(5)))

        x = np.sin(theta) * np.cos(phi)
        y = np.sin(theta) * np.sin(phi)
        z = np.cos(theta)
        
        errs = np.empty_like(x)

        for i in range(x.size):
            
            i_point = nearest_neighbour((x.flat[i], y.flat[i], z.flat[i]), points,
                    max_distance=intp_dist)
            
            if i_point is False:
                errs.flat[i] = 0
            else:
                errs.flat[i] = values[i_point]
        return errs
   
    kdtree = KDTree(points)

    def color_function_optimized(theta, phi):
        
        intp_dist = (2 * np.sin(np.radians(5)))

        x = (np.sin(theta) * np.cos(phi))
        y = (np.sin(theta) * np.sin(phi))
        z = np.cos(theta)
        
        errs = np.empty_like(x)
        positions = [[x.flat[i], y.flat[i], z.flat[i]] for i in range(x.size)]
        
        if used_scipy_version < (1,6,0):
            distances, i_points = kdtree.query( positions, n_jobs=-1 )
        else:
            distances, i_points = kdtree.query( positions, workers=-1 )

        for i in range(errs.size):
            if distances[i] < intp_dist:
                errs.flat[i] = values[i_points[i]]
            else:
                errs.flat[i] = 0
        return errs


    colors = color_function_optimized(theta, phi)
    
    if colormap == 'own':
        culurs = [(0.2, 0.1, 0),(1,0.55,0),(1,1,0.4)]
    elif colormap == 'own-diverge':
        culurs = [(0,(0,0,0)),(0.001,(1,0,0)), (0.5,(1,1,1)), (1,(0,0,1))]
    else:
        # Must be Matplotlib colormap othewise
        culurs = matplotlib.cm.get_cmap(colormap).colors

    ownmap = matplotlib.colors.LinearSegmentedColormap.from_list('ownmap', culurs, 100)
    ax.plot_surface(X, Y, Z, facecolors=ownmap(colors), linewidth=0, vmin=vmin, vmax=vmax)

    
    m = cm.ScalarMappable(cmap=ownmap)
    m.set_array(colors)
    return m    


def histogram_heatmap(all_errors, nbins=20, horizontal=True, drange=None):
    '''

    all_errors      [errors_rot1, errors_rot1, ....] where
                    errors_rot_i = [error1, error2, ...], error_i is float
    '''
    N_bins = 20
    #N_bins = 3
    if drange == 'auto':
        data_range = (np.min(all_errors), np.max(all_errors))
        print('histogram_heatmap data_range {}'.format(data_range))
    elif drange != 'auto':
        data_range = drange
    else:
        data_range = (0, 1)

    image = []

    for rotation_errors in all_errors:
        hist, bin_edges = np.histogram(rotation_errors, bins=N_bins, range=data_range)
        image.append(hist)

    image = np.array(image)

    if horizontal:
        image = image.T

    return image



def save_3d_animation(manalyser, ax=None, plot_function=None,interframe_callback=print,
        i_worker=None, N_workers=None, animation_type='rotate_plot', video_writer=False,
        *args, **kwargs):
    '''
    interframe_callback
    
    manalyser : object or list of objects
        Either one analyser or a list of analysers

    animation_type : string
        "rotate_plot" or "rotate_arrows"
    '''
    try:
        # If manalyer is a list of manalysers
        manalyser[0]
        manalysers = manalyser
    except:
        # If it is actually only one manalyser
        manalysers = [manalyser]
    
    
    fps = 30
    frameskip = False
    
    biphasic=False
    optimal_ranges = []
    
    # FIXME No predetermined optimal ranges
    if len(manalysers) > 1 and animation_type not in['rotate_plot']:
        
        if manalysers[0].manalysers[0].__class__.__name__ == 'MAnalyser':
            optimal_ranges = [[24.3-3, 24.3+3, 'Typical photoreceptor\nmovement axis']]
        elif manalysers[0].__class__.__name__ == 'FAnalyser':
            optimal_ranges = [[-80-3, -80+3, 'Typical ommatidial\nrhabdomere aligment']]
            
            if plot_function.__class__.__name__ == 'plot_3d_vectormap':
                kwargs['arrow_rotations'] = [0, 29]

    if animation_type == 'rotate_plot':
        animation = make_animation_angles(step=0.5 * (20/fps))
    elif animation_type == 'rotate_arrows':
        
        animation = np.linspace(-90, 90, 12*fps)
        
               
    elif animation_type in ['pitch_rot', 'yaw_rot', 'roll_rot']:
        biphasic = True
        animation = np.linspace(-180, 180, 16*fps)
        
        if animation_type == 'pitch_rot':
            optimal_ranges = [[0, 20, 'Typical head tilt\nrange']]
        else:
            optimal_ranges = [[-10, 10, 'Typical head tilt\nrange']]
    
    if len(optimal_ranges) > 0:
        A = optimal_ranges[0][0]
        B = optimal_ranges[-1][1]
        
        for optimal_range in optimal_ranges:
            a,b,n = optimal_range
            start, trash, end = np.split(animation, [np.where(a<=animation)[0][0], np.where(b<animation)[0][0]] )
            
            animation = np.concatenate( (start, np.linspace(a, b, 3*fps), end) )

 


    if i_worker is None:
        partname = ''
    else:
        partname = '_' + str(i_worker)
        
        worksize = math.floor(len(animation) / N_workers)

        if i_worker == N_workers-1:
            animation = animation[i_worker*worksize:]
        else:
            animation = animation[i_worker*worksize:(i_worker+1)*worksize]
        
        # Fast forward to right animation timestep
        for i_frame_before in range(i_worker*worksize):
            make_animation_timestep(step_size=0.05*(20/fps), twoway=biphasic)

    if plot_function:
        
        kwargs['animation_variable'] = animation[0]
        kwargs['animation_type'] = animation_type
        kwargs['animation'] = animation
        kwargs['optimal_ranges'] = optimal_ranges
        kwargs['pulsation_length'] = CURRENT_ARROW_LENGTH
        
        po = plot_function(*manalysers, i_frame=0, *args, **kwargs)
        ax = po[0]
        
        try:
            len(ax)
            axes = ax
        except TypeError:
            axes = [ax]

   
    title = plot_function.__name__ + '_{}_'.format(animation_type) + '_'.join([ma.manalysers[0].__class__.__name__ for ma in manalysers])
    
    savedir = os.path.join(ANALYSES_SAVEDIR, 'videos', title+'_'+str(datetime.datetime.now()))
    os.makedirs(savedir, exist_ok=True)


    if video_writer:
        doublegrab_next = False
        try:
            video_writer = matplotlib.animation.writers['ffmpeg'](fps=fps, metadata={'title': title}, codec='mjpeg')
            video_writer.setup(axes[0].figure, os.path.join(savedir,'{}{}.avi'.format(title ,partname)), dpi=200)
        except RuntimeError:
            print('Install ffmpeg by "pip install ffmpeg" to get the video')
            video_writer = False

    

    for i, animation_variable in enumerate(animation[1:]):
        
        if frameskip:
            if i % frameskip != 0:
                continue

        if i_worker is None:
            i_frame = i
        else:
            i_frame = i + i_worker * worksize

        #try:
        
        if callable(plot_function):
            
            kwargs['animation_variable'] = animation_variable
            kwargs['animation_type'] = animation_type
            kwargs['animation'] = animation
            kwargs['optimal_ranges'] = optimal_ranges
            kwargs['pulsation_length'] = CURRENT_ARROW_LENGTH

            for ax in axes:
                ax.clear()
            make_animation_timestep(step_size=0.05*(20/fps), twoway=biphasic)
            
            if len(axes) > 1:
                plot_function(*manalysers, i_frame=i_frame, axes=axes, *args, **kwargs)
            else:
                plot_function(*manalysers, i_frame=i_frame, ax=axes[0], *args, **kwargs)
        
        print('Animation variable: {}'.format(animation_variable))
        
        if animation_type == 'rotate_plot':
            for ax in axes:
                ax.view_init(elev=animation_variable[0], azim=animation_variable[1])
        
        axes[0].figure.canvas.draw_idle()
        interframe_callback()

        if video_writer:
            video_writer.grab_frame()
            if doublegrab_next:
                video_writer.grab_frame()
                doublegrab_next = False
        
        ax.dist=7
        axes[0].figure.savefig(os.path.join(savedir, 'frame_{0:07d}.png'.format(i_frame)), transparent=True, dpi=300)
        #except Exception as e:
        #    print('Could not make a frame, error message on the next line')
        #    print(e)
        #    doublegrab_next = True

    if video_writer:
        video_writer.finish()





import numpy as np
from gonioanalysis.drosom.kinematics import magstd_over_repeats, mean_max_response


def response_magnitude(manalysers, group_name='noname',
        imagefolder_endswith=''):
    '''
    
    - group name
    - condition/stimulus name (imagefolder_endswith)
    - mean response amplitude
    - variation between animals
    - mean variation in between animals
    - the number N of animals
    '''
    
    cols = []

    respmags = []
    respstds = []

    for manalyser in manalysers:
        image_folders = manalyser.list_imagefolders(endswith=imagefolder_endswith, only_measured=True)
        
        for image_folder in image_folders:
            respmags.append( mean_max_response(manalyser, image_folder, maxmethod='mean_latterhalf') )
            respstds.append( magstd_over_repeats(manalyser, image_folder, maxmethod='mean_latterhalf') )
        

    cols.append(group_name)
    cols.append(imagefolder_endswith)
    cols.append(np.mean(respmags))
    cols.append(np.std(respmags))
    cols.append(np.mean(respstds))
    cols.append(len(manalysers))

    return cols


def response_magnitudes(grouped_manalysers, stimuli = ['uv', 'green']):
    '''
    Statistics for manalyser groups.
    See resposne_magnitude

    Arguments
    ---------
    grouped_manalysers : dict of lists of objects
        Keys are group names, items are lists that contain the manalyser
        objects.
    '''
    rows = []

    for name, manalysers in grouped_manalysers.items():
        for stimulus in stimuli:
            rows.append( response_magnitude(manalysers, group_name=name,
                    imagefolder_endswith=stimulus) )

    
    for row in rows:
        print(row)

    return rows






'''Export displacement timeseries.
'''

import os

import numpy as np

from gonioanalysis.directories import ANALYSES_SAVEDIR
from gonioanalysis.drosom.kinematics import (
        magstd_over_repeats,
        sigmoidal_fit,
        )
from .left_right import write_CSV_cols


SAVEDIR = os.path.join(ANALYSES_SAVEDIR, 'displacement_exports')


def mean_folder_repeats(manalysers, group_name, wanted_imagefolders=None,
                        microns=False,
                        savedir=SAVEDIR):
    '''Takes the mean of the displacements in an image folder.

    If multiple sampling frequencies or recording lengths are present,
    splits them into corresponding CSV files.
    '''
    
    traces = {}

    for manalyser in manalysers:

        if wanted_imagefolders:
            _image_folders = wanted_imagefolders.get(manalyser.name, [])
        else:
            _image_folders = manalysers.list_imagefolders()

        for image_folder in _image_folders:
 
            trace = manalyser.get_magnitude_traces(
                    None, image_folder, mean_repeats=True,
                    microns=microns)
            trace = list(trace.values())
            if len(trace) != 1:
                raise RuntimeError
            trace = trace[0]
            trace = trace.tolist()[0]
            
            # Create a key from the sampling freq and amount of points
            # X-axes are different for these so have to separate
            fs = manalyser.get_imaging_frequency(image_folder)
            points = len(trace)
            key = f'{fs}-{points}'

            if key not in traces:
                traces[key] = []

            trace.insert(0, f'{manalyser.name};{image_folder}')
            traces[key].append(trace)


    os.makedirs(savedir, exist_ok=True)

    for key, trace in traces.items():
        fs, points = key.split('-')

        # Create xaxis
        N = int(points)
        xaxis = np.linspace(0, N/float(fs), N).tolist()
        xaxis.insert(0, 'Time (s)')
        trace.insert(0, xaxis)

        # Add the mean trace in the end
        onlydata = [t[1:] for t in trace[1:]]
        mean = np.mean(onlydata, axis=0).tolist()
        mean.insert(0, 'Mean')
        trace.append(mean)

        if len(traces) != 1:
            fn = os.path.join(savedir, f'{group_name}_{fs}Hz_{points}p.csv')
        else:
            fn = os.path.join(savedir, f'{group_name}.csv')
        write_CSV_cols(fn, trace)




def mean_parallel_repeats(manalysers, group_name, wanted_imagefolders=None,
        savedir=SAVEDIR):
    '''
    Here we mean the repeat 1 of all flies together, then
    the repeat 2 of all flies together, and so on.

    This is for example for the intensity series where the stimulus intensity
    increases by every repeat and we want to know the mean response
    of the flies to the flash (repeat) 1, the flash (repeat) 2, and so on.
    '''

    all_traces = []

    for manalyser in manalysers:

        if wanted_imagefolders:
            _image_folders = wanted_imagefolders.get(manalyser.name, [])
        else:
            _image_folders = manalyser.list_imagefolders()
        
        for image_folder in _image_folders:
            
            for eye in manalyser.eyes:
                traces = manalyser.get_magnitude_traces(eye,
                        image_folder=image_folder)

                traces = list(traces.values())
                if traces:
                    all_traces.append(traces[0])

    # Average repeat 1 together, repeat 2 together etc.
    mean_traces = []

    for i_repeat in range(len(all_traces[0])):
        mean = np.mean([data[i_repeat] for data in all_traces], axis=0)
        mean_traces.append(mean.tolist())
    
    os.makedirs(savedir, exist_ok=True)
    write_CSV_cols(os.path.join(savedir, group_name+'.csv'), mean_traces)



def parallel_repeat_stds(manalysers, group_name, wanted_imagefolders=None,
        savedir=SAVEDIR):
    '''
    Variation within an specimen(s) (not between specimens)
    '''
    stds = [['name', 'disp-std', 'speed-std', '1/2-time std']]
    for manalyser in manalysers:
        if wanted_imagefolders:
            _image_folders = wanted_imagefolders.get(manalyser.name, [])
        else:
            _image_folders = manalyser.list_imagefolders()
        
        for image_folder in _image_folders:
            
            std = [np.std(z) for z in sigmoidal_fit(manalyser, image_folder)]
            
            std[0] = magstd_over_repeats(manalyser, image_folder, maxmethod='mean_latterhalf')
            
            std.insert(0, manalyser.name+'_'+image_folder)

            stds.append(std)


    os.makedirs(savedir, exist_ok=True)
    write_CSV_cols(os.path.join(savedir, group_name+'.csv'), stds)




'''
A script to make an all-in-one Windows installer for gonio-analyis using
pip, pynsist and NSIS.

You have to
    - be on the Windows platfrom (64-bit)
    - have pynsist and NSIS installed
    - have the same major.minor version of Python as PYTHONVERSION in here

Because the wheels embedded in the installer are fetched from PyPi, this script
can make installers only for PyPi released versions of gonio-analysis.

Attributes
----------
GONIOVERSION : string
    Version of the Gonio Analysis to use.
    By default, it is assumed that we are in a git work copy.
PYTHONVERSION : string
    Version of the Python interpreter to use
'''

import os
import sys
import shutil

try:
    # Version number to __version__ variable
    exec(open("../gonioanalysis/version.py").read())
except:
    __version__ == input('gonioanalysis version use (exmpl. 0.1.2) >>')

GONIOVERSION = __version__
PYTHONVERSION = '{}.{}.{}'.format(*sys.version_info[0:3])

def fetch_wheels():
    
    if os.path.isdir('wheels'):
        shutil.rmtree('wheels')
    os.makedirs('wheels')
    
    os.chdir('wheels')
    os.system('pip download gonio-analysis=='+GONIOVERSION)
    os.chdir('..')
    


def build(gonioversion, pythonversion):
    
    os.system('get_tkdeps.py')
    
    fetch_wheels()
    wheels = [os.path.join('wheels', fn) for fn in os.listdir('wheels') if fn.endswith('.whl')]
    
    str_wheels = '\n '.join(wheels)

    moveversion = [fn for fn in wheels if 'movemeter-' in fn][0]
    moveversion = moveversion.split('-')[1]
    
    cfg_file = []
    with open('pynsist_template.cfg', 'r') as fp:
        for line in fp:
            edited = line.replace('GONIOVERSION', gonioversion)
            edited = edited.replace('PYTHONVERSION', pythonversion)
            edited = edited.replace('LOCAL_WHEELS', str_wheels)
            edited = edited.replace('MOVEVERSION', moveversion)
            cfg_file.append(edited)

    with open('pynsist_temp.cfg', 'w') as fp:
        for line in cfg_file:
            fp.write(line)

    print(cfg_file)

    os.system('pynsist pynsist_temp.cfg')
    #shutil.rmtree('wheels')

def main():
    build(GONIOVERSION, PYTHONVERSION)

if __name__ == "__main__":
    main()




'''
Exporting results for the set of experiments where
one location in each left/right eye was measured. 
'''
import os
import csv
import numpy as np

from gonioanalysis.drosom.kinematics import (
        mean_max_response,
        _sigmoidal_fit,
        _simple_latencies,
        )
from gonioanalysis.directories import ANALYSES_SAVEDIR

LR_SAVEDIR = os.path.join(ANALYSES_SAVEDIR, 'LR_exports')


def write_CSV_cols(fn, columns):
    '''
    Note
    -----
    Writes as many rows as there are rows in the first columns.

    Attributes
    ----------
    columns : list of lists
        Each item is column, and in each item that is also a list,
        each item is the value of the row.
    '''


    with open(fn, 'w') as fp:
        writer = csv.writer(fp, delimiter=',')
        for i in range(len(columns[0])):
            row = []
            for sublist in columns:
                try:
                    row.append(sublist[i])
                except:
                    row.append('')
            writer.writerow(row)
            #writer.writerow([sublist[i] for sublist in columns])


def read_CSV_cols(fn):
    rows = []
    with open(fn, 'r') as fp:
        reader = csv.reader(fp, delimiter=',')
        for row in reader:
            rows.append(row)

    return list(map(list, zip(*rows)))


def left_right_displacements(manalysers, group_name,
        fn_prefix='LR-displacements',
        savedir=LR_SAVEDIR,
        stimuli={'uv': ['uv', ')'], 'green': ['green'], 'NA': []},
        strong_weak_division=False, divide_threshold=3,
        wanted_imagefolders=None,
        microns=True, phase=False, mean_lr=False,
        reference_frame=False):
    '''
    Saves CSV files of left and right eye movements and ERGs.
    
    If many recordings for an eye/stimulus/specimen combination exist,
    then takes the mean of these (so that each eye appears only once).


    Arguments
    ----------
    manalysers : list of objects
        MAnalyser objects for
    group_name : string
        Name that describes the manalysers. For example, "blind_norpa" or
        "controls".
    fn_prefix : string
        Text to append in the beginnign of the CSV filename.
    stimuli : dict of lists of strings
        Each key is the name of the stimulus, and matching value is a list of
        the suffixes that match the stimulus (the suffix in the end of imagefolder name)
    strong_weak_division : bool
        If True, group data based on strong and weak eye instead of
        combined left and right.
    divide_threshold : int
        Related to the strong_weak_divison argument. For some specimen there may be
        recordings only from one eye, and divide_threshold or more is required
        to in total to do the division.
    wanted_imagefolders: None or a dict
        Keys specimen names, items a sequence of wanted imagefolders
        Relaxes horizontal conditions.
    microns : bool
        Convert pixel movement values to microns
    phase : bool
        If True, return phase (vector direction) instead of the magnitude.
    mean_lr : bool
        If True, average the left and right eye data together.
    reference_frame : False or int
        If an interger (between 0 and N_frames-1), use the corresponding
        frame as a reference zero point.

    '''
    
    # each "file" is a list of columns
    
    fs = None
    efs = None
    
    if wanted_imagefolders:
        conditions = [None, None] 
        csv_files = {'NA': []}
    else:
        conditions = [lambda h: h>10, lambda h: h<-10]
        csv_files = {stim: [] for stim in stimuli.keys()}

    for manalyser in manalysers:
        
        # Left eye
        for eye, condition in zip(['left', 'right'], conditions):
            
            if eye=="left" or mean_lr == False:
                eyedata = {stim: [] for stim in stimuli.keys()}

            for image_folder in manalyser.list_imagefolders(horizontal_condition=condition): 
                
                if wanted_imagefolders and image_folder not in wanted_imagefolders.get(manalyser.name, []):
                    # Skip image_folder if not in wanted_imagefolders (if it's not None)
                    continue
                
                if wanted_imagefolders is None:
                    # look for match
                    stims = []
                    for _stim in stimuli.keys():
                        if any([image_folder.endswith(match) for match in stimuli[_stim]]):
                            stims.append( _stim )
                else:
                    stims = ['NA']
                
                for stim in stims:

                    trace = manalyser.get_magnitude_traces(eye,
                            image_folder=image_folder,
                            mean_repeats=True, microns=microns,
                            _phase=phase)
                
                    trace = list(trace.values())
                    if len(trace) >= 2:
                        raise NotImplementedError('mistake in implementation')
                    

                    if trace:
                        # Check that fs matches
                        nfs = manalyser.get_imaging_frequency(image_folder=image_folder)
                        if fs is None:
                            fs = nfs
                        elif fs != nfs:
                            raise ValueError('Analysers with multiple fs!')

                        trace = trace[0][0]
                        
                        if reference_frame is not False:
                            trace = [val - trace[reference_frame] for val in trace]

                        eyedata[stim].append(trace)
       
            if eye == "right" or mean_lr == False:
                for stim in stimuli.keys():
                    if eyedata[stim]:
                        column_name = '{}_mean_{}'.format(manalyser.name, eye)

                        csv_files[stim].append( np.mean(eyedata[stim], axis=0).tolist() )
                        csv_files[stim][-1].insert(0, column_name)


        if "ERGs" in manalyser.linked_data:
            data = manalyser.linked_data['ERGs']
            
            repeatdata = {}

            erg_columns = {}
            
            for recording in data:
                name = 'ERGs_' + recording['Stimulus'] +'!;!'+ recording['eye']
                try:
                    N_repeats = int(recording['N_repeats'])
                except ValueError:
                    N_repeats = 1

                if repeatdata.get(name, 0) < N_repeats: 
                    erg_columns[name] = (np.array(recording['data'])-recording['data'][0]).tolist()
                    erg_columns[name].insert(0, '{}_{} (mV)'.format(manalyser.name, recording['eye']))
            
                    if efs is None:
                        efs = recording['fs']
                    elif efs != recording['fs']:
                        raise ValueError('ERGs with multiple sampling frequencies!')

            for name in erg_columns:
                uname = name.split('!;!')[0]
                try:
                    csv_files[uname]
                except:
                    csv_files[uname] = []

                csv_files[uname].append(erg_columns[name])
    
    
    if strong_weak_division:
        new_csv_files = {}
    
        # Process first DPP data then ERGs
        strong_eyes = {}
        keys = [k for k in csv_files if not 'ERGs' in k] + [k for k in csv_files if 'ERGs' in k]        

        for csv_file in keys:
            pairs = []

            column_titles = [column[0] for column in csv_files[csv_file]]

            for column in csv_files[csv_file]:
                
                if not 'right' in column[0]:
                    try:
                        indx = column_titles.index( column[0].replace('left', 'right'))
                    except ValueError:
                        continue

                    pairs.append((column, csv_files[csv_file][indx]))

           
            if len(pairs) > divide_threshold:
                new_csv_files[csv_file+'_strong'] = []
                new_csv_files[csv_file+'_weak'] = []

                for left, right in pairs:
                    # Fixme
                    rdata = [float(num) for num in right[1:]]
                    ldata = [float(num) for num in left[1:]]

                    if not 'ERGs' in csv_file:
                        specimen_name = '_'.join(left[0].split('_')[:-2])
                    else:
                        specimen_name = '_'.join(left[0].split('_')[:-1])
                    
                    print(specimen_name)

                    if 'ERGs' in csv_file:
                        #ab = [400, 800]
                        
                        if strong_eyes[specimen_name] == 'right':
                            new_csv_files[csv_file+'_strong'].append(right)
                            new_csv_files[csv_file+'_weak'].append(left)
                        else:
                            new_csv_files[csv_file+'_strong'].append(left)
                            new_csv_files[csv_file+'_weak'].append(right)
                    else:
                        ab = [None, None]
                        if abs(quantify_metric(rdata, ab=ab)) > abs(quantify_metric(ldata, ab=ab)):
                            new_csv_files[csv_file+'_strong'].append(right)
                            new_csv_files[csv_file+'_weak'].append(left)
                            strong_eyes[specimen_name] = 'right'
                        else:
                            new_csv_files[csv_file+'_strong'].append(left)
                            new_csv_files[csv_file+'_weak'].append(right)
                            strong_eyes[specimen_name] = 'left'

            else:
                new_csv_files[csv_file+'_all'] = csv_files[csv_file]
            
        csv_files = new_csv_files



    os.makedirs(savedir, exist_ok=True)

    for csv_file in csv_files:
        # Mean in the end
        csv_files[csv_file].append(np.mean([csv_files[csv_file][i][1:] for i in range(len(csv_files[csv_file]))], axis=0).tolist())
        try:
            csv_files[csv_file][-1].insert(0, 'mean')
        except AttributeError as e:
            if csv_file.startswith('ERGs'):
                print(csv_files[csv_file])
                raise ValueError("No ERGs, check linking the ERG data")
            else:
                #raise e
                continue

        if csv_file.startswith('ERGs_'):
            ufs = efs
        else:
            ufs = fs

        # Add xaxis (time) in all files
        data = csv_files[csv_file][0][1:]
        xaxis = np.linspace(0, (len(data)-1)/ufs, len(data)).tolist()
        xaxis.insert(0, 'time (s)')
        csv_files[csv_file].insert(0, xaxis)

        fn = '{}_{}_{}.csv'.format(fn_prefix, group_name, csv_file)
        fn = os.path.join(savedir, fn)
        write_CSV_cols(fn, csv_files[csv_file])


def quantify_metric(data1d, metric_type='mean', ab=(None, None)):
    '''
    From a 1D array (time series) quantify single value metric.

    metric_type : string
        "mean" to take the mean of the range
    ab : tuple of integers
        The range as datapoint indices.
    '''
    part = data1d
    if ab[1] is not None:
        part = part[:ab[1]]
    if ab[0] is not None:
        part = part[ab[0]:]

   
    if metric_type == 'mean':
        value = np.mean(part)

    return value


def lrfiles_summarise(lrfiles, point_type='mean', ab=(None, None)):
    '''
    Datapoints for making box/bar plots and/or for statistical testing.

    Arguments
    ---------
    lrfiles : list of filenames
        LR-displacements files of the left_right_displacements.
    point_type : string
        Either "mean" to take mean of the range (used for DPP movement data) or
        min-start to take the mean around the minimum and subtract start value (used for ERGs).
        If not specified, use 'mean'.
    ab : tuple
        Specify the range as indices (rows of the lrfiles excluding the header)
        If not specified, try to autodetect based on if ERGs is contained in the filesames
        ((half,end) for DPP, (400, 800) for ERGs).
    '''
    
    csv_files = {}
    

    for fn in lrfiles:
        
        
        sfn = os.path.basename(fn)
        specimen_name = '_'.join(sfn.split('_')[1:-1])
        stim = sfn.split('_')[-1].split('.')[0]

        csv_rows = csv_files.get(stim, {})

        
        coldata = read_CSV_cols(fn)
        
        if specimen_name not in csv_rows:
            csv_rows[specimen_name] = []
        
        if ab[0] is None or ab[1] is None:
        
            # FIXME Quite specific
            if 'ERGs' in sfn:
                a, b = [400, 800]
            else:
                a, b = [int((len(coldata[0])-1)/2), len(coldata[0])-1]
        else:
            a, b = ab

        # First column is time, the last is the mean, skip these
        for col in coldata[1:-1]:
            
            if point_type != 'kinematics':
                if a is not None and b is not None:
                    numerical_col = [float(num) for num in col[a+1:b]]
                else:
                    numerical_col = [float(num) for num in col[1:]]
            else:
                numerical_col = [float(num) for num in col[1:]]

            if point_type == 'mean':
                value = np.mean(numerical_col)
            elif point_type.startswith('min-start'):
                value = np.min(numerical_col) - float(col[1])
            
            elif point_type == 'kinematics':
                fs = 1/(float(coldata[0][2]) - float(coldata[0][1]))
                value = _sigmoidal_fit([numerical_col], fs)
                
                if value is None:
                    continue

                value = [value[0][0], value[1][0], value[2][0]]
                #value = _simple_latencies([numerical_col], fs)[0]
            
            if len(lrfiles)==1 and point_type == "kinematics":
                # expand CSV rows
                for name, val in zip(['displacement', 'logistic growth rate', '1/2-risetime'], value):
                    try:
                        csv_rows[specimen_name+'_'+name]
                    except:
                        csv_rows[specimen_name+'_'+name] = []
                    csv_rows[specimen_name+'_'+name].append(val)

            else:
                csv_rows[specimen_name].append(value)


        csv_files[stim] = csv_rows

    path = os.path.join(os.path.dirname(lrfiles[0]), 'summary')
    os.makedirs(path, exist_ok=True)

    for fn in csv_files:
        ofn = os.path.join( path, 'LR_summary_{}.csv'.format(fn) )
        with open(ofn, 'w') as fp:
            writer = csv.writer(fp, delimiter=',')
            for row in csv_files[fn]:
                writer.writerow([row]+csv_files[fn][row])



def left_right_summary(manalysers):
    '''
    Condensed from left/right
    '''
    csvfile = []
    
    header = ['Specimen name', 'Left mean response (pixels)', 'Right mean response (pixels)', 'Right eye ERG response (mV)', 'Eyes with microsaccades', '"Normal" ERGs']
    
    csvfile.append(header)

    for manalyser in manalysers:
        
        print(manalyser.get_specimen_name())

        line = []

        line.append(manalyser.get_specimen_name())
        
        # Left eye

        responses = []
        
        for image_folder in manalyser.list_imagefolders(horizontal_condition=lambda h: h>=0): 
            if image_folder.endswith("green"):
                continue
            
            print(image_folder)
            mean = mean_max_response(manalyser, image_folder)
            if not np.isnan(mean):
                responses.append(mean)
        
        line.append(np.mean(responses))
        
        # Right eye
        responses = []
        
        for image_folder in manalyser.list_imagefolders(horizontal_condition=lambda h: h<0): 
            if image_folder.endswith('green'):
                continue
            mean = mean_max_response(manalyser, image_folder)
            if not np.isnan(mean):
                responses.append(mean)
        
        line.append(np.mean(responses))
        
   
        uvresp = None
        greenresp = None
        
        print(manalyser.linked_data.keys())

        if "ERGs" in manalyser.linked_data:
            data = manalyser.linked_data['ERGs']
            
            print(data[0].keys())

            for recording in data:
                print(recording.keys())
                if int(recording['N_repeats']) == 25:

                    if recording['Stimulus'] == 'uv':
                        uvresp = np.array(recording['data'])
                    if recording['Stimulus'] == 'green':
                        greenresp = np.array(recording['data'])

        

        uvresp = uvresp - uvresp[0]
        uvresp = np.mean(uvresp[500:1000])
        
        greenresp = greenresp - greenresp[0]
        greenresp = np.mean(greenresp[500:1000])
        
        line.append(uvresp)
        
        # Descide groups
        dpp_threshold = 0.9999
        if line[1] < dpp_threshold and line[2] < dpp_threshold:
            line.append('None')
        elif line[1] < dpp_threshold or line[2] < dpp_threshold:
            line.append('One')
        else:
            line.append('Two')

        erg_threshold = 0.5
        if abs(line[3]) < erg_threshold:
            line.append('No')
        else:
            line.append('Yes')


        csvfile.append(line)


    for line in csvfile:
        print(line)
    
    with open("left_right_summary.csv", 'w') as fp:
        writer = csv.writer(fp)
        for line in csvfile:
            writer.writerow(line)


    







import numpy as np


from .analysing import AnalyserBase


def insert_virt_suffix(folder_name, i_virt):
    parts = folder_name.split('_')
    parts.insert(1, f'virt{i_virt}')
    return '_'.join(parts)

def remove_virt_suffix(folder_name):
    parts = folder_name.split('_')
    if parts[1].startswith('virt'):
        i_virt = int(parts.pop(1).removeprefix('virt'))
    return '_'.join(parts), i_virt

class VirtualAnalyser(AnalyserBase):
    '''Creates a new virtual analyser based on one or more analysers

    Similar to MAverager but does not average and interpolate
    data.

    Each image folder gets a virt suffix in its name to avoid
    name collisions. For example, "pos(0,0)_bigsmoke" becomes
    "pos(0,0)_virt0_bigsmoke" (and "pos(0,0)_virt1_bigsmoke" if
    the second analyser also has this exact same folder)
    '''

    def __init__(self, name, analysers):
        super().__init__()

        self.name = name
        self.folder = name

        self.analysers = analysers
        self.eyes = analysers[0].eyes

        self.vector_rotation = 0
        self.attributes = {'yaw': 0}

    @property
    def active_analysis(self):
        return self.analysers[0].active_analysis

    @active_analysis.setter
    def active_analysis(self, name):
        for an in self.analysers:
            an.active_analysis = name


    def list_analyses(self):
        analyses = []
        for an in self.analysers:
            for analysis in an.list_analyses():
                if analysis in analyses:
                    continue
                analyses.append(analysis)
        return analyses
    
    def save_attributes(self):
        pass

    def mark_bad(self, image_folder, i_repeat, i_is_relative=True):
        folder, i_virt = remove_virt_suffix(image_folder)
        return self.analysers[i_virt].mark_bad(
                folder, i_repeat, i_is_relative=i_is_relative)

    # List rotations is missing, not sure how to handle this
    # and if it would be better to remove list_rotations alltogether
    #def list_rotations(self)
    
    def list_imagefolders(self, *args, **kwargs):
        folders = []
        for i_an, an in enumerate(self.analysers):
            fols = an.list_imagefolders(*args, **kwargs)

            # Add index of the analyser to the folder suffix beginning
            fols = [insert_virt_suffix(fol, i_an) for fol in fols]

            folders.extend(fols)
        
        return folders

    def get_horizontal_vertical(self, image_folder, *args, **kwargs):
        folder, i_virt = remove_virt_suffix(image_folder)
        return self.analysers[i_virt].get_horizontal_vertical(
                folder, *args, **kwargs)

    def get_specimen_directory(self):
        '''Returns a list of the directories
        '''
        return [an.get_specimen_directory() for an in self.analysers]

    def get_N_repeats(self, image_folder, *args, **kwargs):
        folder, i_virt = remove_virt_suffix(image_folder)
        return self.analysers[i_virt].get_N_repeats(
                folder, *args, **kwargs)

    def list_images(self, image_folder, *args, **kwargs):
        folder, i_virt = remove_virt_suffix(image_folder)
        return self.analysers[i_virt].list_images(
                folder, *args, **kwargs)

    def get_specimen_name(self):
        return self.name
   
    def get_imaging_parameters(self, image_folder):
        folder, i_virt = remove_virt_suffix(image_folder)
        return self.analysers[i_virt].get_imaging_parameters(folder)

    def get_specimen_age(self):
        '''Assumes all specimens have the same age
        '''
        ages = [an.get_specimen_age() for an in self.analysers]
        # FIXME: add warning if ages do not match
        return ages[0]

    def get_specimen_sex(self):
        '''Assumes all specimens have the same sex
        '''
        sexes = [an.get_specimen_sex() for an in self.analysers]
        # FIXME: add warning if sexes do not match
        return sexes[0]
    
    def get_imaging_frequency(self, image_folder, fallback_value=100):
        folder, i_virt = remove_virt_suffix(image_folder)
        return self.analysers[i_virt].get_imaging_frequency(
                folder, fallback_value)

    def get_pixel_size(self, image_folder):
        folder, i_virt = remove_virt_suffix(image_folder)
        return self.analysers[i_virt].get_pixel_size(folder)

    def get_snap_fn(self, i_snap=0, absolute_path=True):
        # FIXME: Only gives snaps from the first specimen
        return self.analysers[0].get_snap_fn(
                i_snap, absolute_path=absolute_path)
    
    def load_ROIs(self):
        for an in self.analysers:
            an.load_ROIs()

    def select_ROIs(self, **kwargs):
        for an in self.analysers:
            an.select_ROIs(**kwargs)

    def are_rois_selected(self, *args, **kwargs):
        selected = [an.are_rois_selected() for an in self.analysers]
        return all(selected)
    
    def count_roi_selected_folders(self):
        count = 0
        for an in self.analysers:
            count = an.count_roi_selected_folders()
        return count

    def folder_has_rois(self, image_folder):
        '''
        image_folder : string
            Name of the image folder with the virt suffix
        '''
        folder, i_virt = remove_virt_suffix(image_folder)
        return self.analysers[i_virt].folder_has_rois(folder)

    def get_rois(self, image_folder):
        folder, i_virt = remove_virt_suffix(image_folder)
        return self.analysers[i_virt].get_rois(folder)

    def is_measured(self, *args, **kwargs):
        measured = [an.is_measured() for an in self.analysers]
        return all(measured)
    
    def folder_has_movements(self, image_folder):
        folder, i_virt = remove_virt_suffix(image_folder)
        return self.analysers[i_virt].folder_has_movements(folder)

    def load_analysed_movements(self, *args, **kwargs):
        for an in self.analysers:
            an.load_analysed_movements(*args, **kwargs)

    def get_recording_time(image_folder, i_rep=0):
        return None

    def measure_both_eyes(self, **kwargs):
        for an in self.analysers:
            an.measure_both_eyes(**kwargs)

    def measure_movement(self, eye, *args, **kwargs):
        for an in self.analysers:
            an.measure_movement(eye, *args, **kwargs)
   
    def get_time_ordered(self, *args, **kwargs):
        # FIXME: Doesn not check which eye was measured first
        # FIXME: May break some logic that uses this method?
        image_fns = []
        ROIs = []
        angles = []
        for an in self.analysers:
            fns, rois, angs = an.get_time_ordered(*args, **kwargs)
            image_fns.extend(fns)
            ROIs.extend(rois)
            angles.extend(angs)
        return image_fns, ROIs, angles

    def get_movements_from_folder(self, image_folder):
        folder, i_virt = remove_virt_suffix(image_folder)
        return self.analysers[i_virt].get_movements_from_folder(folder)

    def get_displacements_from_folder(self, image_folder):
        folder, i_virt = remove_virt_suffix(image_folder)
        return self.analysers[i_virt].get_displacements_from_folder(
                folder)
   
    def get_2d_vectors(self, eye, *args, **kwargs):
        # FIXME: Probably doesn't give sense making results
        # with yaw combining.
        angles = []
        X = []
        Y = []
        for an in self.analysers:
            ang, x, y = an.get_2d_vectors(eye, *args, **kwargs)
            angles.extend(ang)
            X.extend(x)
            Y.extend(y)
        return angles, X, Y

    def get_magnitude_traces(self, eye, image_folder=None,
                             *args, **kwargs):
        
        # Case A: image_folder is given
        if image_folder is not None:
            folder, i_virt = remove_virt_suffix(image_folder)
            return self.analysers[i_virt].get_magnitude_traces(
                    eye, folder, *args, **kwargs)
        
        # Case B: image_folder is None
        # -> append suffix virt_ian to all keys
        traces = {}
        for i_an, an in enumerate(self.analysers):
            data = an.get_magnitude_traces(eye, None, *args, **kwargs)
            for key, value in data.items():
                traces[f'{key}_virt{i_an}'] = value

        return traces

    def get_moving_ROIs(self, eye, image_folder, i_repeat=0):
        folder, i_virt = remove_virt_suffix(image_folder)
        return self.analysers[i_virt].get_moving_ROIs(
                eye, folder, i_repeat=i_repeat)
        

    def get_3d_vectors(self, eye, *args, **kwargs):
        
        points = []
        vectors = []

        for analyser in self.analysers:
            
            yaw = analyser.attributes['yaw']
            
            # Make sure the coloring comes up correct
            # yaw == 0: default case, manalyser code splits L-R
            # yaw==90 or yaw==-90: Should be only one eye recording
            if yaw == 0:
                P, V = analyser.get_3d_vectors(eye, *args, **kwargs)
                points.append(P)
                vectors.append(V)
            elif (yaw == 90 and eye == 'left') or (yaw == -90 and eye == 'right'):
                for aeye in analyser.eyes:
                    P, V = analyser.get_3d_vectors(
                            aeye, *args, **kwargs) 
                    if len(P) > 0:
                        points.append(P)
                        vectors.append(V)

        points = np.concatenate((*points,))
        vectors = np.concatenate((*vectors,))

        merge_distance = kwargs.get('merge_distance', 0)
        if merge_distance:
            points, vectors = self._merge_by_distance(
                    points, vectors, merge_distance)

        return points, vectors

    
    def get_recording_time(self, image_folder, *args, **kwargs):
        folder, i_virt = remove_virt_suffix(image_folder)
        return self.analysers[i_virt].get_recording_time(
                folder, *args, **kwargs)
    
    def stop(self):
        for an in self.analysers:
            an.stop()





import numpy as np
import tifffile
import matplotlib.pyplot as plt


def open_adjusted(fn):
    '''
    Images can have high glares. This function tries to
    clip the image (adjust contrast) so that the DPP
    is clearly visible.

    Usually the glare is at one point so this uses
    the fact.
    '''

    image = tifffile.imread(fn)

    glare_max = np.max(image)
    median = np.median(image)

    image = np.clip(image, 0, int(np.mean([glare_max, median])/6) )

    return image



class ImageShower:
    '''
    From  cascade training, a generalized image shower
    '''
    
    def __init__(self, fig, ax):
        
        self.fig = fig
        self.ax = ax
        
        self.fns = None

        self.buttonActions = []
        self.image_brightness = 0
        self.image_maxval = 1
        self.cid = self.fig.canvas.mpl_connect('key_press_event', self.callbackButtonPressed)
        
        self.pupils = -1
        
        self.title = ''

        self.fig.show()


    def callbackButtonPressed(self, event):
        '''
        A callback function connecting to matplotlib's event manager.
        '''
        
        # Navigating between the images
        if event.key == 'z':
            self.image_maxval -= 0.3
            self.updateImage(strong=True)
        
        elif event.key == 'x':
            self.image_maxval += 0.3
            self.updateImage(strong=True)
        
        elif event.key == 'a':
            self.image_brightness += 50
            self.updateImage(strong=True)
        elif event.key == 'c':
            self.image_brightness += -50
            self.updateImage(strong=True)
        
        for button, action in self.buttonActions:
            if event.key == button:
                print(event.key)
                action()


    def setImages(self, image_fns):
        '''
        Set the images that ImageShower shows.
        '''
        self.fns = image_fns
        self.cache = {}

    def setImage(self, i):
        fn = self.fns[i]
        try:
            self.image = self.cache[fn]
        except KeyError:
            self.image = tifffile.imread(fn) 
        
        self.updateImage()
        
    def cacheImage(self, i):
        '''
        Loads image to cache for faster showup. 
        '''
        fn = self.fns[i]
        # Uncomment to take cahcing in use
        #self.cache[fn] = tifffile.imread(fn)

    def setTitle(self, title):
        self.title = title

    def updateImage(self, strong=False):
        capvals = (0, np.mean(self.image) *self.image_maxval)
        self.ax.clear() 
        self.ax.set_title(self.title)
        self.ax.imshow(self.image-self.image_brightness,cmap='gist_gray', interpolation='nearest', vmin=capvals[0], vmax=capvals[1])
        
        if not strong:
            self.fig.canvas.draw()
        else:
            self.fig.show()
    
    def close(self):
        plt.close(self.fig)





import os
import json
import warnings
import csv

import matplotlib.pyplot as plt
import numpy as np

from gonioanalysis.directories import PROCESSING_TEMPDIR, ANALYSES_SAVEDIR
from gonioanalysis.rotary_encoders import to_degrees


def load_angle_pairs(fn):
    '''
    Loading angle pairs from a file.
    
    Detached from gonio_imsoft.
    '''
    angles = []
    with open(fn, 'r') as fp:
        reader = csv.reader(fp)
        for row in reader:
            if row:
                angles.append([int(a) for a in row])
    return angles


def load_data(folder, arl_fly=False):
    '''
    Loading a data folder.

    Returns a list where horizontal angles are grouped by the pitch angle
    and each horizontal angle is next to it's image's filename.

    grouped = [pit1, [[hor1, fn1], ...] ...]
    

    INPUT ARGUMENTS     DESCRIPTION
    arl_fly             Set true if normal DrosoX processing should be skipped
                            (meaning no outliner remove, pitch grouping etc...)

    '''
    
    if os.path.isdir(os.path.join(folder, 'rot')):
        # DrosoX format (MM + trigger): Sequence saved images + anglepairs.txt
        fns = [os.path.join(folder,'rot',fn) for fn in os.listdir(os.path.join(folder, 'rot')) if fn.endswith('.tif')]
        fns.sort()
        
        angles = load_angle_pairs(os.path.join(folder, 'anglepairs.txt'))
        
        # FIXME: Cannot really load any stack to check how many images in it,
        #       takes too long if remote filesystem
        # If saved as stack (much less images than angles)
        if 10*len(fns) < len(angles):
            fns = [fns[0]+'_{}'.format(i) for i in range(len(angles))]


    else:
        # DrosoM format (gonio imsoft): each position in own folder
        # Here for DrosoX, use only the first image in each pos folder
        # if many exists
        _folders = [f for f in os.listdir(folder) if f.startswith('pos')]
        fns = [[os.path.join(folder, f, fn) for fn in
            os.listdir(os.path.join(folder, f)) if fn.endswith('.tiff')][0] for f in _folders]
        
        angles = [ [int(n) for n in f.split('(')[1].split(')')[0].split(',')] for f in _folders]

    to_degrees(angles)
    
    print('Angles {} and images {}'.format(len(angles), len(fns)))
    if abs(len(angles) - len(fns)) > 10:
        warnings.warn("Large missmatch between the number of recorded the angles and images.", UserWarning)
    
    fns, angles = _makeSameLength(fns, angles)
   
    angles_and_images = _pitchGroupedHorizontalsAndImages(fns, angles, arl_fly=arl_fly)
    if not arl_fly:
        
        print('Determing pitches to be combined...')
        angles_to_combine = _pitchesToBeCombined(angles_and_images, angles)
        

        # GROUP NEAR PITCHES
        print('Combining pitches...')
        angles_and_images = _groupPitchesNew(angles_and_images, angles_to_combine) 
        print('After grouping: {}'.format(len(angles_and_images)))

        # NO PROBLEM AFTER THIS
        # -------------------------
        print('Removeing lonely outliners...')
        angles_and_images = _removeOutliners(angles_and_images, 2)

        #angles_and_images = self.removeShorts(angles_and_images)
        
        # SORT PITCHES
        angles_and_images.sort(key=lambda x: x[0], reverse=True)
        
       
        # SORT HORIZONTALS
        for i in range(len(angles_and_images)):
            angles_and_images[i][1].sort(key=lambda x: x[0])
        
    return angles_and_images


def _removeOutliners(angles_and_images, degrees_threshold):
    '''

    '''

    for pitch, hor_im in angles_and_images:
        remove_indices = []

        for i in range(len(hor_im)):
            center = hor_im[i][0]
            try:
                previous = hor_im[i-1][0]
            except IndexError:
                previous = None
            try:
                forward = hor_im[i+1][0]
            except IndexError:
                forward = None

            if not (previous == None and forward == None):

                if forward == None:
                    if abs(previous-center) > degrees_threshold:
                        remove_indices.append(i)
                if previous == None:
                    if abs(forward-center) > degrees_threshold:
                        remove_indices.append(i)

            #if previous != None and forward != None:
            #    if abs(previous-center) > degrees_threshold and abs(forward-center) > degrees_threshold:
            #        remove_indices.append(i)
    
        for i in sorted(remove_indices, reverse=True):
            hor_im.pop(i)

    return angles_and_images


def _getPitchIndex(pitch, angles_and_images):

    for i in range(len(angles_and_images)):
        if angles_and_images[i][0] == pitch:
            return i
    print('Warning: No pitch {} in angles_and_images'.format(pitch))
    return None


def _groupPitchesNew(angles_and_images, to_combine):
    '''
    Rotatory encoders have some uncertainty so that the pitch can "flip"
    to the next value if encoder's position in
    '''
    grouped = []

    for pitches in to_combine:
        
        combinated = []
        for pitch in pitches:
            index = _getPitchIndex(pitch, angles_and_images)

            combinated.extend(angles_and_images.pop(index)[1])
        
        grouped.append([np.mean(pitches), combinated])
        
    
    angles_and_images.extend(grouped)
    
    return angles_and_images


def _makeSameLength(lista, listb):
    if len(lista) > len(listb):
        lista = lista[0:len(listb)]
    elif len(lista) < len(listb):
        listb = listb[0:len(lista)]
    return lista, listb


def _pitchesToBeCombined(angles_and_images, angles):
    '''
    Assuming gonio scanning was done keeping pitch constant while
    varying horizontal angle, it's better to group line scans together
    because there may be slight drift in the pitch angle.
    '''
    
    pitches = [[]]
    scan_direction = -10
    
    anglefied_angles_and_images = []
    for pitch, hor_im in angles_and_images:
        for horizontal, fn in hor_im:
            anglefied_angles_and_images.append([horizontal, pitch])

    # Determine pitches that should be combined
    for i in range(1, len(angles)-1):
        if angles[i] in anglefied_angles_and_images:

            direction = np.sign( angles[i][0] - angles[i-1][0] )
            future_direction = np.sign(angles[i+1][0] - angles[i][0])
            
            if direction != scan_direction and not future_direction == scan_direction:
                pitches.append([])
                scan_direction = direction
           
            if direction == scan_direction or (scan_direction == 0 and future_direction == scan_direction):
                if not angles[i][1] in pitches[-1]:
                    pitches[-1].append(angles[i][1])

    
    pitches = [p for p in pitches if len(p)>=2 and len(p)<5]
    
   
    # A pitch can appear more than one time to be combined. This seems
    # usually happen in adjacent pitch groupings.
    # Here, combine [a,b] [b,c] -> [a,b,c]
    combine = []
    for i in range(len(pitches)-1):
        for j, pitch in enumerate(pitches[i]):
            if pitch in pitches[i+1]:                    
                combine.append([i, j])

    for i, j in sorted(combine, reverse=True):
        pitches[i].pop(j)
        pitches[i] += pitches[i+1]
        pitches.pop(i+1)
    # ----------------------------------------------------- 

    print("Pitches to be combined")
    for p in pitches:
        print(p)

    return pitches


def _pitchGroupedHorizontalsAndImages(image_fns, angles, arl_fly=False):
    '''
    Returns horizontal angles grouped by pitch (as groupHorizontals)
    but also links image fn with each horizontal angle.
    
    Note: image_fns and angles must have one to one correspondence.
    
    IDEAL STRUCTURE TO WORK WITH

    grouped = [pit1, [[hor1, fn1], ...] ...]
    '''

    grouped = []
    pitches_in_grouped = []
    
    for fn, (horizontal, pitch) in zip(image_fns, angles):
        
        if not pitch in pitches_in_grouped:
            pitches_in_grouped.append(pitch)
            grouped.append([pitch, []])
        
        i = pitches_in_grouped.index(pitch)
        grouped[i][1].append([horizontal, fn])
    
    # For each pitch angle there must be more than 10 images
    # or the whole pitch is removed
    if not arl_fly:
        grouped = [x for x in grouped if len(x[1]) > 10]
    else:
        print('ARL fly, not capping of imaging row.')

    return grouped




import os
import argparse

import matplotlib.pyplot as plt

from gonioanalysis.droso import DrosoSelect, simple_select
from gonioanalysis.drosox.analysing import XAnalyser
from gonioanalysis.drosox.plotting import (
        plot_1d_overlap,
        plot_matrix_overlap,
        plot_experiment_illustration
)


def main():

    parser = argparse.ArgumentParser(description='DrosoX: Analyse DPP static imaging data')
    parser.add_argument('--datadir',
            help='Path to the specimens data directory or ASK')
    
    parser.add_argument('--specimens',
            help='Space separated list of specimens. Ased if not specified.',
            nargs='*')

   
    parser.add_argument('--measure-overlap',
            help='(Re)measure binocular overlap using user-assisted binary search',
            action='store_true')

    
    # Overlap data
    parser.add_argument('--print-overlap',
            help='Print overlap data as text',
            action='store_true')

    # Plotting arguments
    parser.add_argument('--plot-overlap',
            help='Plot 1D binocular overlap',
            action='store_true')
    parser.add_argument('--plot-matrix-overlap',
            help='Plot binocular overlap as discrete rectangulars',
            action='store_true')
    parser.add_argument('--plot-experiment-illustration',
            help='Plot illustrative video how the experiments were done',
            action='store_true')



    args = parser.parse_args()
    
    if args.datadir.lower() == 'ask':
        datadir = input("Input data directory >> ")
    else:
        datadir = args.datadir


    if not args.specimens:
        specimens = DrosoSelect(datadir=datadir).ask_user()
        specimens = [os.path.basename(specimen) for specimen in specimens]
    else:
        specimens = args.specimens
    

    xanalysers = [XAnalyser(datadir, specimen) for specimen in specimens]

    # Perform the tasks that have to be performed per xanalyser
    for xanalyser in xanalysers:
        print(xanalyser.fly)
        if args.print_overlap:
            xanalyser.print_overlap()

        if args.measure_overlap:
            xanalyser.measure_overlap()
        
        if args.plot_experiment_illustration:
            plot_experiment_illustration(xanalyser)

   
    if args.plot_overlap:        
        plot_1d_overlap(xanalysers)
    elif args.plot_matrix_overlap:
        plot_matrix_overlap(xanalysers)

    if not args.measure_overlap:
        plt.show()


if __name__ == "__main__":
    main()




'''
Extension of find antenna levels to create a reference fly.

This script has to be (ideally) run only once on DrosoALR flies, creating
an antenna lvl reference fly that can be used for other flies as well.
'''

import os

import numpy as np
import tifffile

from gonioanalysis.antenna_level import AntennaLevelFinder
from gonioanalysis.directories import PROCESSING_TEMPDIR_BIGFILES

from imalyser.averaging import Templater
from imalyser.aligning import Aligner
from imalyser.common import imwrite


DROSO_DATADIR = input('Input data directory >> ')


def loadReferenceFly(folder):
    '''
    Returns the reference fly data, dictionary with pitch angles as keys and
    image filenames as items.
    '''
    pitches = []
    with open(os.path.join(folder, 'pitch_angles.txt'), 'r') as fp:
        for line in fp:
            pitches.append(line)

    images = [os.path.join(folder, fn) for fno in s.listdir(folder) if fn.endswith('.tif') or fn.endswith('.tiff')]
    
    return {pitch: fn for pitch,fn in zip(pitches, fns)}


class ReferenceCreator:
    '''
    Create a refenrence fly that can be loaded like an ALR (antenna level reference) fly.
    '''
    
    def __init__(self, name):
        self.name = name
        self.savedir = 'alr_data'
        os.makedirs(self.savedir, exist_ok=True)

    def _loadData(self):
        '''
        Loads all the present ALR flies.
        '''
        
        reference_data = []

        
        alr_flies = [fn for fn in os.listdir(DROSO_DATADIR) if 'DrosoALR' in fn] 
        alr_folders = [os.path.join(DROSO_DATADIR, fn)for fn in alr_flies] 

        lfinder = AntennaLevelFinder()
        
        for folder,fly in zip(alr_folders, alr_flies):    
            reference_data.append( lfinder._load_drosox_reference(folder, fly) )
        
        return reference_data



    def _restructureData(self, reference_data):
        '''
        Restructures data from self._loadData to work in self.createReferenceFly.
        Limits that are not present in all the flies are removed (FIXME structure this phrase better).
        
        Input data =  [fly1_dict, fly2_dict, ...] where
                flyi_dict = {image_fn_1: pitch_1, ...}

        Output data = {ptich_1: [image_fn_1_fly1,...]}
        '''
        restructured_data = {}

        step_size = 1 # degree
        N = len(reference_data)
        
        # 1) MAKE ALL FLIES TO SPAN THE SAME ANGLES
        angles = [[angle for image_fn, angle in reference_data[i].items()] for i in range(N)]

        mins = [np.min(angles[i]) for i in range(N)]
        maxs = [np.max(angles[i]) for i in range(N)]

        limits = (int(np.max(mins)), int(np.min(maxs)))



        # 2)  BIN together based on step_size
        
        for iter_pitch in range(*limits, step_size):
            
            restructured_data[str(iter_pitch)] = []

            for fly_i_data in reference_data:
                for image_fn, pitch in fly_i_data.items():
                    
                    if iter_pitch <= pitch < iter_pitch + step_size:
                        restructured_data[str(iter_pitch)].append( image_fn )
            
        return restructured_data

    def createReferenceFly(self):
        
        data = self._loadData()
        data = self._restructureData(data)

        templater = Templater()
        aligner = Aligner()

        templates = []
        angles_order = []

        for i, (angle, images) in enumerate(sorted(data.items(), key=lambda x:float(x[0]))):
            
            print('Templating for pitch {} degrees ({}/{}), {} images to align and average together.'.format(angle, i+1, len(data), len(images)))

            template = templater.template(images)
            templates.append(template)
            #imwrite(fn, template)

            angles_order.append(angle)

        print('Aligning all template images...')
        offsets = aligner.calcOffsets(templates)
        templates = aligner.getAligned(templates, offsets)
        
        for i, template in enumerate(templates):
            

            fn = os.path.join(self.savedir, 'im{:0>5}.tif'.format(i))
            print('Saving {}'.format(fn))
            tifffile.imwrite(fn, template)

        with open(os.path.join(self.savedir, 'pitch_angles.txt'), 'w') as fp:
            for pitch in angles_order:
                fp.write(pitch+'\n')




def main():
    
    creator = ReferenceCreator('Drosophila')
    creator.createReferenceFly()

if __name__ == "__main__":
    main()



'''
Goniometric Analysis
---------------------

Python pacakge `gonioanalysis` provides an intergrated set of graphical
and command line tools to analyse goniometric imaging data.
Here goniometry means that the rotation of a sample with respect
to the imaging device is well documented while rotating the sample
between the imaging runs.

More precesily, `gonioanalysis` takes advantage of the following data structuring

    data_directory
    ├── specimen_01
    │   ├── pos(horizontal, vertical)_some-suffix-stufff
    │   │   ├── image_rep0_0.tif
    │   │   └── image_rep0_1.tif
    │   └── ...
    └── ...

Here, the rotation of the sample is encoded in the image folder name
with <em>pos</em> prefix.
Alternatively, images with a generalized directory structure can be also used

    data_directory
    ├── specimen_01
    │   ├── experiment_01
    │   │   ├── image_001.tif
    │   │   └── image_002.tif
    │   └── ...
    └── ...


Please mind that `gonioanalysis` was created to analyse <em>Drosophila</em>'s
deep pseudogonio movement and orientation data across its left and right eyes
and because of this the language used in many places can be droso-centric.


User interfaces
---------------

Currently `gonioanalysis` comes with two user interfaces, one being command line based and
the other graphical.

The documentation on this page gives out only general instrctions and ideas.
To have exact usage instructions, please refer to submodule documentation of `gonioanalysis.drosom.terminal` and
`gonioanalysis.tkgui`.


Initial analysis
----------------
The following two steps have to be performed for every imaged specimen, but luckily only once.
The first one is manual needing user intercation but can be performed quite fast depending the amount of image folders,
The second one is automated but can take a long time depending on the amount of data.

### 1) Selecting regions of interest (ROIs)

In `gonioanalysis`, rectangular ROIs are manually drawn by the user
once per each image folder.
The idea is to confine the moving feature inside a rectangle but only in the first frame. 


### 2) Measuring movements

After the ROIs have been selected, movement analysis can be run.
Using 2D cross-correlation based methods (template matching), `gonioanalysis`
automatically quntifies the movement of the ROI bound method across all the frames and repeats.

Because this part requires no user interactions and takes a while,
usually it is best to select all the ROIs for all specimens beforehand,
then batch run the movement analysis.


After initial analysis
----------------------

### Vectormap - 3D directions

The 2D data of the moving features in the camera coordinates can be transferred into a 3D coordinate
system of specimens frame of reference.
To do so, we need to have the rotation of the specimen with respect to the camera specified
in some fixed coordinate system, most naturally in the coordinate system set by the rotation stages
that are used to rotate the fly.
At the moment, `gonioanalysis` supports only one rotation stage configuration that is
a specimen fixed on a vertical rotation stage that is fixed on a horizontal rotation stage that is fixed on a table.


### Orientation analysis
To analyse directionaly of any arbitrary features (such as hair pointin directions across the head),
you can override the `gonioanalysis.drosom.analysing.MAnalyser` with `gonioanalysis.drosom.orientation_analysis.OAnalyser`,
making the movement measurement part to be a manual user drawing arrows according to the feature direction analysis.
Then later, you can use any `gonioanalysis` code paths just remembering that for example in the vectormap,
the arrows point the direction of the features, not their movement.


Exporting data
--------------

All quitely saved files are stored in a location set in `gonioanalysis.directories`.
By default, on Windows this is <em>C:/Users/USER/GonioAnalysis</em>
and on other platforms <em>/home/USER/.gonioanalysis</em>


Command line interface
----------------------

`gonioanalysis` also includes a command interface, that can be invoked by

    python -m gonioanalysis.drosom.terminal


For all different options and help use `--help` option.
In case no ROIs have been selected, the 
When elections of the ROIs cannot be done in headless environments.
'''

from .version import __version__



'''
Estimating optic flow field
'''

from math import cos, sin, radians

import numpy as np
import scipy
from scipy.spatial import cKDTree as KDTree
from scipy.stats import mannwhitneyu

import gonioanalysis.coordinates as coordinates
from gonioanalysis.drosom.analysing import MAnalyser
from gonioanalysis.version import used_scipy_version


def flow_direction(point, xrot=0):
    '''
    Estimates optic flow at the given point by placing the optic flow vector to
    the point, and forcing it to the tangent plane of the sphere.

    The optic flow vector is a unit vector -j (ie. pointing to negative y-axis),
    unless rotated. This should be okay since we are interested only in the
    direction of the flow field, not magnitude.

    INPUT ARGUMENTS         DESCRIPTION
    point                   (x0,y0,z0)
    xrot                    Rotation about x-axis

    RETURNS
    xi,yj,zk    Flow direction vector at origo
    '''
    
    rxrot = radians(xrot)
    ov = 2*np.array(point) + np.array([0,-1*cos(rxrot),sin(rxrot)])

    P1 = coordinates.force_to_tplane(point, ov)
    
    P1 = coordinates.normalize(point, P1, scale=0.10)

    return P1-np.array(point)



def flow_vectors(points, xrot=0):
    '''
    Returns optic flow vectors (from flow_direction) as a numpy array.
    '''
    return np.array([flow_direction(P0, xrot=xrot) for P0 in points])


def field_error(points_A, vectors_A, points_B, vectors_B, direction=False, colinear=False):
    '''
    Relieved version where points dont have to overlap
    
    Put to A the eye vecotrs

    vectors_X   list of vector
    vector      (x,y,z)

    direction   Try to get the direction also (neg/pos)

    colinear : bool
    
    Returns the errors at points_A
    '''
    
    N_vectors = len(vectors_A)

    errors = np.empty(N_vectors)

    kdtree = KDTree(points_B)
    
    if used_scipy_version < (1,6,0):
        distances, indices = kdtree.query(points_A, k=10, n_jobs=-1)
    else:
        distances, indices = kdtree.query(points_A, k=10, workers=-1)
    weights = 1/(np.array(distances)**2)
    
    # Check for any inf
    for i_weights in range(weights.shape[0]):
        if any(np.isinf(weights[i_weights])):
            weights[i_weights] = np.isinf(weights[i_weights]).astype('int')

    #if any(np.isinf(weights):
    compare_vectors = [[vectors_B[i] for i in indx] for indx in indices]

    for i, (vecA, vecBs, vecB_weights) in enumerate(zip(vectors_A, compare_vectors, weights)):
        
        vec_errors = []
        
        for vecB in vecBs:

            angle = np.arccos(np.inner(vecA, vecB)/(np.linalg.norm(vecA) * np.linalg.norm(vecB)))
            error = angle / np.pi
            if not 0<=error<=1:
                # Error is nan if either of the vectors is zero because this leads to division
                # by zero because np.linalg.norm(vec0) = 0
                # -> set error to 1 if vecA != vecB or 0 otherwise
                if np.array_equal(vecA, vecB):
                    error = 0
                else:
                    error = 1
            
            if direction:
                counter = coordinates.rotate_along_arbitrary(points_A[i], vecB, angle) 
                clock = coordinates.rotate_along_arbitrary(points_A[i], vecB, -angle)
                
                if np.sum(counter - vecB) > np.sum(clock - vecB):
                    error = -error

            vec_errors.append(error)

        errors[i] = np.average(vec_errors, weights=vecB_weights)
    
    if direction:
        if colinear:
            errors *= 2
        errors = (errors + 1)/2
    else:
        if colinear:
            errors = 2 * np.abs(errors - 0.5)
        else:
            errors = 1 - errors

    return errors


def _find_unique_points(points):
    '''
    Returns
    -------
    points : list
    indices : dict of int
    '''

    unique_points = list({p for p in points})
    indices = [[] for i in range(unique_points)]
    for i_point, point in enumerate(points):
        indices[unique_points.index(point)].append(i_point)
    
    return unique_points, indices


def _angle(self, vecA, vecB):
    return np.arccos(np.inner(vecA, vecB)/(np.linalg.norm(vecA) * np.linalg.norm(vecB)))

def field_pvals(points_A, vectors_A, points_B, vectors_B, direction=False, colinear=False):
    '''
    Assuming 
    '''
    # same points are repeated many times (otherwise it would make sense
    # to do statistical testing. Find the these "unique" points
    un_points_A, un_indices_A = _find_unique_points(points_A)
    un_points_B, un_indices_B = _find_unique_points(points_B)

    kdtree = KDTree(points_B)
    
    for point, indices_A in zip(un_points_A, un_indices_A):
        # Closest point
        if used_scipy_version < (1,6,0):
            distance_B, index_B = kdtree.query(point, k=1, n_jobs=-1)
        else:
            distance_B, index_B = kdtree.query(point, k=1, workers=-1)
        
        Avecs = [vectors_A[i] for i in indices_A]
        Bvecs = [vectors_B[i] for i in un_indices_B[index_B]]
        
        # Mean of Avecs
        mean_Avec = np.mean(Avecs, axis=0)
        
        # Relative rotations of vectors with respect to the mean Avec
        Adirs = [_angle(vec, mean_Avec) for vec in Avecs]
        Bdirs = [_angle(vec, mean_Avec) for vec in Avecs]

        u_stats, pval = mannwhitneyu(Adirs, Bdirs)
        
        pvals.append(pval)

    return un_points_A, pvals



class FAnalyser(MAnalyser):
    '''
    Sham analyser to just output optic flow vectors with the same
    api as MAnalyer does.
    '''
    
    def __init__(self, *args, **kwargs):
            
        print('inited')
        
        self.manalysers = [self]
        self.folder = 'optic_flow'
        self.eyes = ['left', 'right']
        self.vector_rotation = 0
        

        # FAnalyser specific
        self.pitch_rot = 0
        self.roll_rot = 0
        self.yaw_rot = 0
        self.points = {'right': coordinates.optimal_sampling(np.arange(0, 60, 5), np.arange(-100, 100, 5)),
                'left': coordinates.optimal_sampling(np.arange(-60, 0, 5), np.arange(-100, 100, 5))}
        
        self.constant_points = False
        

        self.ui_options = {
                'pitch_rot': {'help': 'Pitch rotation', 'type': float},
                'roll_rot': {'help': 'Roll rotation', 'type': float},
                'yaw_rot': {'help': 'Yaw rotation', 'type': float},
                }


    def get_3d_vectors(self, eye, constant_points=None, *args, **kwargs):
        '''
        
        constant_points : bool
            If true, points stay the same and only vectors get rotated.
            If false, smooth rotation of the whole optic flow sphere.
        '''

        if constant_points is None:
            constant_points = self.constant_points

        if constant_points:
            # Rotate points, calculate vectors, rotate back
            points = coordinates.rotate_points(self.points[eye],
                    radians(self.yaw_rot),
                    radians(self.pitch_rot),
                    radians(self.roll_rot))
            
            points, vectors = coordinates.rotate_vectors(points, flow_vectors(points, xrot=0),
                    -radians(self.yaw_rot),
                    -radians(self.pitch_rot),
                    -radians(self.roll_rot))
        else:
            points = coordinates.optimal_sampling(np.arange(-90,90,5), np.arange(-180,180,5))
            points, vectors = coordinates.rotate_vectors(points, flow_vectors(points, xrot=0),
                    -radians(self.yaw_rot),
                    -radians(self.pitch_rot),
                    -radians(self.roll_rot))
            
            # Fixme. Make me with numpy, not list comprehension
            if eye == 'left':
                indices = [i for i, point in enumerate(points) if point[0] <= 0]
            elif eye == 'right':
                indices = [i for i, point in enumerate(points) if point[0] >= 0]

            points = [points[i] for i in indices]
            vectors = [vectors[i] for i in indices]
        
        return np.array(points), np.array(vectors)

    
    def is_measured(self, *args, **kwargs):
        return True

    def are_rois_selected(self, *args, **kwargs):
        return True

    def load_analysed_movements(self, *args, **kwargs):
        return None






# Bring all from common to this name space
from .common import *

# But from the "big widgets" only the needed class(es)
from .zero_correct import ZeroCorrect
from .compare_vectormaps import CompareVectormaps
from .export_wizard import ExportWizard
